Package: mx.api
Type: Package
Title: Minimal Matrix Client-Server API
Version: 0.3.0
Date: 2026-06-10
Authors@R: c(
    person("Troy", "Hernandez", role = c("aut", "cre"),
           email = "troy@cornball.ai",
           comment = c(ORCID = "0009-0005-4248-604X")),
    person("cornball.ai", role = "cph"))
Description: A minimal-dependency client for the 'Matrix' Client-Server
    HTTP API <https://spec.matrix.org/>, suitable for talking to a
    'Synapse' <https://element-hq.github.io/synapse/> or 'Conduit'
    <https://conduit.rs/> homeserver. Covers login, room management,
    message send and history, media upload or download, and the
    transport endpoints needed to coordinate end-to-end encryption
    (device-key and one-time-key publication, key query and claim,
    to-device events). Encryption itself is out of scope;
    pair with a separate crypto package.
License: MIT + file LICENSE
URL: https://github.com/cornball-ai/mx.api
BugReports: https://github.com/cornball-ai/mx.api/issues
Imports: curl, jsonlite
Suggests: tinytest
Encoding: UTF-8
NeedsCompilation: no
Packaged: 2026-06-10 17:50:22 UTC; troy
Author: Troy Hernandez [aut, cre] (ORCID:
    <https://orcid.org/0009-0005-4248-604X>),
  cornball.ai [cph]
Maintainer: Troy Hernandez <troy@cornball.ai>
Repository: CRAN
Date/Publication: 2026-06-10 18:00:02 UTC

# tinyrox says don't edit this manually, but it can't stop you!

export(mx_canonical_json)
export(mx_delete_device)
export(mx_devices)
export(mx_download)
export(mx_get_account_data)
export(mx_get_state)
export(mx_guess_mime)
export(mx_keys_claim)
export(mx_keys_query)
export(mx_keys_upload)
export(mx_login)
export(mx_logout)
export(mx_media_config)
export(mx_messages)
export(mx_profile)
export(mx_react)
export(mx_read_receipt)
export(mx_redact)
export(mx_register)
export(mx_room_create)
export(mx_room_invite)
export(mx_room_join)
export(mx_room_leave)
export(mx_room_members)
export(mx_room_name)
export(mx_room_topic)
export(mx_rooms)
export(mx_send)
export(mx_send_audio)
export(mx_send_event)
export(mx_send_file)
export(mx_send_image)
export(mx_send_media)
export(mx_send_to_device)
export(mx_send_video)
export(mx_session)
export(mx_set_account_data)
export(mx_set_avatar_url)
export(mx_set_displayname)
export(mx_set_state)
export(mx_sync)
export(mx_typing)
export(mx_upload)
export(mx_whoami)

# mx.api

A minimal-dependency client for the
[Matrix Client-Server HTTP API](https://spec.matrix.org/), suitable for
talking to a Synapse or Conduit homeserver from R. Two Imports: `curl`
and `jsonlite`. No tidyverse.

Pairs with [`mx.crypto`](https://github.com/cornball-ai/mx.crypto),
which handles Olm + Megolm; mx.api itself does no cryptography.

## Install

```r
# CRAN
install.packages("mx.api")

# GitHub (development version)
remotes::install_github("cornball-ai/mx.api")
```

## Quick start: log in, send, sync

```r
library(mx.api)

s <- mx_login(
    server   = "https://matrix.example",
    user     = "alice",
    password = "hunter2"
)

room <- mx_room_join(s, "#general:matrix.example")
mx_send(s, room, "hello from R")

batch <- mx_sync(s, timeout = 0)
str(batch$rooms$join)

mx_logout(s)
```

## Create a private room with invites

```r
room <- mx_room_create(
    s,
    name   = "project sync",
    topic  = "weekly check-in",
    preset = "private_chat",
    invite = c("@bob:matrix.example", "@carol:matrix.example")
)
mx_send(s, room, "kickoff in 5")
```

`mx_room_create` returns the new room id as a character string.
Presets: `"private_chat"`, `"trusted_private_chat"`, `"public_chat"`.

## Sending rich-text (code blocks, bold, etc.)

`mx_send`'s `body` is plain text. Matrix clients that show rich text
(Element, Cinny, Fractal, etc.) render the optional `formatted_body`
field instead, fall back to `body` when it is absent. Markdown
fences in plain `body` show as literal backticks; you have to send
HTML in `formatted_body` to get a real code block.

Pass both via the `extra` argument:

```r
plain <- "Build steps:\n\ndocker compose build\ndocker compose up -d"

html <- paste0(
    "<p>Build steps:</p>",
    "<pre><code>docker compose build\n",
    "docker compose up -d</code></pre>"
)

mx_send(
    s, room, plain,
    extra = list(
        format = "org.matrix.custom.html",
        formatted_body = html
    )
)
```

The `format` value `"org.matrix.custom.html"` is the only one the
Matrix spec defines. Supported tags are listed in the
[Matrix client-server spec, section 11.2.1.7](https://spec.matrix.org/v1.11/client-server-api/#mroommessage-msgtypes).
Code blocks use `<pre><code>…</code></pre>`; inline code is `<code>…</code>`.

## What's covered

| Area | Functions |
|---|---|
| Session | `mx_register`, `mx_login`, `mx_logout`, `mx_whoami`, `mx_session` |
| Rooms | `mx_rooms`, `mx_room_create`, `mx_room_join`, `mx_room_leave`, `mx_room_invite`, `mx_room_members`, `mx_room_name`, `mx_room_topic` |
| Messages | `mx_send`, `mx_send_event`, `mx_messages`, `mx_sync`, `mx_react`, `mx_redact`, `mx_read_receipt`, `mx_typing` |
| Room state | `mx_get_state`, `mx_set_state` |
| Media | `mx_upload` (streaming), `mx_download`, `mx_send_media` (+ `mx_send_file` / `mx_send_image` / `mx_send_audio` / `mx_send_video`), `mx_guess_mime`, `mx_media_config` |
| Profile | `mx_profile`, `mx_set_displayname`, `mx_set_avatar_url` |
| Account data | `mx_get_account_data`, `mx_set_account_data` |
| Devices | `mx_devices`, `mx_delete_device` |
| E2EE transport | `mx_keys_upload`, `mx_keys_query`, `mx_keys_claim`, `mx_send_to_device` |
| E2EE signing helper | `mx_canonical_json` |

End-to-end **cryptography** is out of scope; pair with `mx.crypto`
(or another crypto library) to sign and verify the payloads these
endpoints carry. Helpful framing:

- mx.api speaks Matrix HTTP. It does no signing, no key management,
  no key validation, and holds no state between calls.
- mx.crypto speaks Olm + Megolm. It does no HTTP.
- [mx.client](https://github.com/cornball-ai/mx.client) is the stateful
  layer that uses both: config persistence, sync cursors, and E2EE
  orchestration (its `vignette("e2ee")` walks the whole flow).

## Canonical JSON

`mx_canonical_json()` is the byte-stable encoder Matrix's signing rules
require. It is hand-rolled (not a jsonlite wrapper) so the
spec-sensitive choices — key ordering by UTF-8 byte sequence, integer
range, NaN/Inf/NA rejection, duplicate-key rejection, control-char
escaping — are visible and unit-tested rather than hidden in another
package's defaults.

```r
mx_canonical_json(list(b = 2, a = 1))
#> [1] "{\"a\":1,\"b\":2}"

mx_canonical_json(1.5)
#> Error: mx_canonical_json: non-integer number 1.5 disallowed
```

107 assertions exercise the encoder (see `inst/tinytest/test_canonical_json.R`).

## Status

**0.3.0**. Relative to 0.2.0 the delta is additive:

- Generic event and state plumbing: `mx_send_event`, `mx_set_state`,
  `mx_get_state` (needed for `m.room.encrypted` / `m.room.encryption`).
- Media messages: `mx_send_media` and the file/image/audio/video
  wrappers.
- Bot lifecycle: `mx_room_invite`, `mx_redact`, `mx_typing`,
  `mx_profile` / `mx_set_displayname` / `mx_set_avatar_url`.
- Account data: `mx_get_account_data`, `mx_set_account_data`.
- Devices: `mx_devices`, `mx_delete_device`.
- Classed error conditions (`mx_error_<ERRCODE>`) for programmatic
  handling; streamed uploads.

See `NEWS.md` for the full changelog.

## CI

GitHub Actions via [r-ci](https://github.com/eddelbuettel/r-ci); macOS
and Ubuntu runners cover every commit + PR.

## License

MIT. See `LICENSE`.

# Room operations

#' List rooms the user has joined
#'
#' @param session An "mx_session" object.
#'
#' @return Character vector of room IDs.
#' @examples
#' \dontrun{
#' mx_rooms(s)
#' }
#' @export
mx_rooms <- function(session) {
    resp <- mx_http(
                    session$server, "GET", "/_matrix/client/v3/joined_rooms",
                    token = session$token
    )
    unlist(resp$joined_rooms, use.names = FALSE)
}

#' Create a room
#'
#' @param session An "mx_session" object.
#' @param name Character or NULL. Human-readable room name.
#' @param topic Character or NULL. Room topic.
#' @param visibility Character. "private" (default) or "public".
#' @param preset Character or NULL. A Matrix room preset
#'   ("private_chat", "trusted_private_chat", "public_chat").
#' @param invite Character vector. Matrix IDs to invite.
#'
#' @return The new room ID as a character string.
#' @examples
#' \dontrun{
#' room_id <- mx_room_create(s, name = "test", topic = "hello")
#' }
#' @export
mx_room_create <- function(session, name = NULL, topic = NULL,
                           visibility = "private", preset = NULL,
                           invite = character()) {
    body <- list(visibility = visibility)
    if (!is.null(name)) {
        body$name <- name
    }
    if (!is.null(topic)) {
        body$topic <- topic
    }
    if (!is.null(preset)) {
        body$preset <- preset
    }
    if (length(invite)) {
        body$invite <- as.list(invite)
    }

    resp <- mx_http(
                    session$server, "POST", "/_matrix/client/v3/createRoom",
                    body = body, token = session$token
    )
    resp$room_id
}

#' Join a room by ID or alias
#'
#' @param session An "mx_session" object.
#' @param room Character. Room ID (!abc:server) or alias (#name:server).
#'
#' @return The joined room ID.
#' @examples
#' \dontrun{
#' mx_room_join(s, "#general:matrix.example")
#' }
#' @export
mx_room_join <- function(session, room) {
    path <- sprintf("/_matrix/client/v3/join/%s", mx_encode_id(room))
    resp <- mx_http(
                    session$server, "POST", path,
                    body = mx_empty_body(), token = session$token
    )
    resp$room_id
}

#' Leave a room
#'
#' @param session An "mx_session" object.
#' @param room_id Character. The room ID.
#'
#' @return Invisible NULL.
#' @examples
#' \dontrun{
#' mx_room_leave(s, "!abc:matrix.example")
#' }
#' @export
mx_room_leave <- function(session, room_id) {
    path <- sprintf("/_matrix/client/v3/rooms/%s/leave", mx_encode_id(room_id))
    mx_http(session$server, "POST", path, body = mx_empty_body(),
            token = session$token)
    invisible(NULL)
}

#' List the members of a room
#'
#' @param session An "mx_session" object.
#' @param room_id Character. The room ID.
#'
#' @return Character vector of Matrix user IDs currently joined.
#' @examples
#' \dontrun{
#' mx_room_members(s, "!abc:matrix.example")
#' }
#' @export
mx_room_members <- function(session, room_id) {
    path <- sprintf("/_matrix/client/v3/rooms/%s/joined_members",
                    mx_encode_id(room_id))
    resp <- mx_http(session$server, "GET", path, token = session$token)
    names(resp$joined)
}

#' Get a room's human-readable name
#'
#' Reads the \code{m.room.name} state event. Returns NULL if the room
#' has no name set or the state event is inaccessible.
#'
#' @param session An "mx_session" object.
#' @param room_id Character. The room ID.
#'
#' @return Character scalar or NULL.
#' @examples
#' \dontrun{
#' mx_room_name(s, "!abc:matrix.example")
#' }
#' @export
mx_room_name <- function(session, room_id) {
    path <- sprintf(
                    "/_matrix/client/v3/rooms/%s/state/m.room.name",
                    mx_encode_id(room_id)
    )
    resp <- tryCatch(
                     mx_http(session$server, "GET", path, token = session$token),
                     error = function(e) NULL
    )
    if (is.null(resp) || is.null(resp$name) || !nzchar(resp$name)) {
        return(NULL)
    }
    resp$name
}

#' Get a room's topic
#'
#' Reads the \code{m.room.topic} state event. Returns NULL if the room
#' has no topic set or the state event is inaccessible.
#'
#' @param session An "mx_session" object.
#' @param room_id Character. The room ID.
#'
#' @return Character scalar or NULL.
#' @examples
#' \dontrun{
#' mx_room_topic(s, "!abc:matrix.example")
#' }
#' @export
mx_room_topic <- function(session, room_id) {
    path <- sprintf(
                    "/_matrix/client/v3/rooms/%s/state/m.room.topic",
                    mx_encode_id(room_id)
    )
    resp <- tryCatch(
                     mx_http(session$server, "GET", path, token = session$token),
                     error = function(e) NULL
    )
    if (is.null(resp) || is.null(resp$topic) || !nzchar(resp$topic)) {
        return(NULL)
    }
    resp$topic
}


#' Invite a user to a room
#'
#' Invitation at creation time is covered by \code{mx_room_create()};
#' this covers the other common lifecycle case, inviting into an
#' existing room.
#'
#' @param session An "mx_session" object.
#' @param room_id Character. The room ID.
#' @param user_id Character. The Matrix ID to invite.
#' @return Invisibly TRUE on success.
#' @examples
#' \dontrun{
#' mx_room_invite(s, "!abc:example", "@friend:example.org")
#' }
#' @export
mx_room_invite <- function(session, room_id, user_id) {
    path <- sprintf(
                    "/_matrix/client/v3/rooms/%s/invite",
                    mx_encode_id(room_id)
    )
    mx_http(session$server, "POST", path,
            body = list(user_id = user_id), token = session$token)
    invisible(TRUE)
}

library(tinytest)

cj <- mx.api::mx_canonical_json

# =========================================================================
# Scalars
# =========================================================================

expect_equal(cj(NULL), "null")
expect_equal(cj(TRUE), "true")
expect_equal(cj(FALSE), "false")
expect_equal(cj(""), "\"\"")
expect_equal(cj("hi"), "\"hi\"")

# =========================================================================
# Numbers — Matrix canonical JSON allows only integers in
# [-(2^53)+1, (2^53)-1]. Float values are explicitly forbidden by the
# spec, so the encoder must REJECT any non-integer numeric input rather
# than serialise it as a decimal.
# =========================================================================

# integer L
expect_equal(cj(0L), "0")
expect_equal(cj(42L), "42")
expect_equal(cj(-7L), "-7")

# numeric (double) that happens to be integer-valued must serialise
# without a decimal place — "1" not "1.0"
expect_equal(cj(0), "0")
expect_equal(cj(1), "1")
expect_equal(cj(1.0), "1")
expect_equal(cj(-1), "-1")
expect_equal(cj(-0), "0")        # spec: "-0 MUST NOT appear" in output

# Non-integer doubles MUST error (spec: "Float values are not permitted
# by this encoding"). Previously these silently produced "1.5", which
# would have signed a non-canonical payload.
expect_error(cj(1.5),  pattern = "non-integer")
expect_error(cj(-1.5), pattern = "non-integer")
expect_error(cj(0.5),  pattern = "non-integer")
expect_error(cj(0.1),  pattern = "non-integer")
expect_error(cj(0.1 + 0.2), pattern = "non-integer")
expect_error(cj(c(1, 1.5)), pattern = "non-integer")  # one bad value in a vector

# Integers within the safe range succeed; values past it must error.
expect_equal(cj(2^53 - 1), "9007199254740991")
expect_equal(cj(-(2^53 - 1)), "-9007199254740991")
expect_error(cj(2^53),    pattern = "out of")
expect_error(cj(-(2^53)), pattern = "out of")
expect_error(cj(1e16),    pattern = "out of")  # 10^16 > 2^53-1

# Integers below 2^53 still must not use scientific notation
expect_false(grepl("e", cj(1e10), fixed = TRUE))
expect_false(grepl("E", cj(1e10), fixed = TRUE))
expect_equal(cj(1e10), "10000000000")
expect_equal(cj(1e15), "1000000000000000")

# rejected number sentinels
expect_error(cj(NA))
expect_error(cj(NA_real_))
expect_error(cj(NA_integer_))
expect_error(cj(NaN))
expect_error(cj(Inf))
expect_error(cj(-Inf))
expect_error(cj(c(1, NaN)))
expect_error(cj(c(1, Inf)))

# rejected character / logical sentinels
expect_error(cj(NA_character_))
expect_error(cj(c("a", NA_character_)))

# =========================================================================
# String escaping
# =========================================================================

# Required JSON escapes
expect_equal(cj("a\"b"), "\"a\\\"b\"")        # quote
expect_equal(cj("a\\b"), "\"a\\\\b\"")        # backslash
expect_equal(cj("a\bb"), "\"a\\bb\"")         # \b (backspace, 0x08)
expect_equal(cj("a\fb"), "\"a\\fb\"")         # \f (form feed, 0x0C)
expect_equal(cj("a\nb"), "\"a\\nb\"")         # \n
expect_equal(cj("a\rb"), "\"a\\rb\"")         # \r
expect_equal(cj("a\tb"), "\"a\\tb\"")         # \t

# Backslash + quote together — easy to break with the wrong gsub order
expect_equal(cj("\\\""), "\"\\\\\\\"\"")      # input: \ "    output: " \ \ \ " "

# Forward slash MUST NOT be escaped (RFC 8259 allows but doesn't require;
# Matrix spec follows the unescaped form, and signers care about exact bytes)
expect_equal(cj("a/b"), "\"a/b\"")
expect_equal(cj("https://matrix.example/path"),
             "\"https://matrix.example/path\"")

# DEL (0x7F) is NOT a JSON control char and must NOT be escaped
expect_equal(cj(intToUtf8(0x7F)),
             paste0("\"", intToUtf8(0x7F), "\""))

# All low control chars 0x01-0x1F escape correctly
expect_equal(cj(intToUtf8(0x01)), "\"\\u0001\"")        # SOH
expect_equal(cj(intToUtf8(0x07)), "\"\\u0007\"")        # BEL
expect_equal(cj(intToUtf8(0x08)), "\"\\b\"")            # BS (named)
expect_equal(cj(intToUtf8(0x09)), "\"\\t\"")            # HT (named)
expect_equal(cj(intToUtf8(0x0A)), "\"\\n\"")            # LF (named)
expect_equal(cj(intToUtf8(0x0B)), "\"\\u000b\"")        # VT
expect_equal(cj(intToUtf8(0x0C)), "\"\\f\"")            # FF (named)
expect_equal(cj(intToUtf8(0x0D)), "\"\\r\"")            # CR (named)
expect_equal(cj(intToUtf8(0x0E)), "\"\\u000e\"")        # SO
expect_equal(cj(intToUtf8(0x1F)), "\"\\u001f\"")        # US

# Multiple control chars in one string, mixed with text
expect_equal(
  cj(paste0("x", intToUtf8(0x01), "y", intToUtf8(0x02), "z")),
  "\"x\\u0001y\\u0002z\""
)

# =========================================================================
# Unicode — non-ASCII passes through as raw UTF-8, no \u escapes
# =========================================================================

# BMP (two-byte UTF-8): "café" with composed é = U+00E9 (UTF-8: C3 A9)
cafe <- "café"
expect_equal(charToRaw(cj(cafe)),
             as.raw(c(0x22, charToRaw(enc2utf8(cafe)), 0x22)))

# BMP (three-byte UTF-8): U+20AC EURO SIGN = E2 82 AC
euro <- intToUtf8(0x20AC)
expect_equal(charToRaw(cj(euro)),
             as.raw(c(0x22, 0xE2, 0x82, 0xAC, 0x22)))

# Supplementary plane (four-byte UTF-8): U+1F980 CRAB = F0 9F A6 80
crab <- intToUtf8(0x1F980)
expect_equal(charToRaw(cj(crab)),
             as.raw(c(0x22, 0xF0, 0x9F, 0xA6, 0x80, 0x22)))

# Combining sequence: "e" + U+0301 COMBINING ACUTE ACCENT (decomposed é)
# Both codepoints must round-trip; we do not normalize (Matrix spec does
# not require NFC).
decomp <- intToUtf8(c(0x65, 0x0301))
expect_equal(charToRaw(cj(decomp)),
             as.raw(c(0x22, charToRaw(enc2utf8(decomp)), 0x22)))

# Mixed ASCII + supplementary in one string
expect_equal(charToRaw(cj(paste0("hi ", crab, "!"))),
             as.raw(c(0x22, charToRaw(enc2utf8("hi ")),
                      0xF0, 0x9F, 0xA6, 0x80,
                      charToRaw(enc2utf8("!")), 0x22)))

# =========================================================================
# Arrays
# =========================================================================

# unnamed lists are arrays
expect_equal(cj(list()), "[]")
expect_equal(cj(list(1L)), "[1]")
expect_equal(cj(list(1L, 2L, 3L)), "[1,2,3]")
expect_equal(cj(list("a", "b")), "[\"a\",\"b\"]")
expect_equal(cj(list(TRUE, FALSE, NULL)), "[true,false,null]")

# length>1 atomic vectors are arrays
expect_equal(cj(c(1L, 2L)), "[1,2]")
expect_equal(cj(c("a", "b")), "[\"a\",\"b\"]")
expect_equal(cj(c(TRUE, FALSE)), "[true,false]")

# length-1 list forces an array (the documented escape hatch)
expect_equal(cj(list(1L)), "[1]")
expect_equal(cj(list("a")), "[\"a\"]")
# vs. length-1 atomic, which is a scalar
expect_equal(cj(1L), "1")
expect_equal(cj("a"), "\"a\"")

# mixed-type array
expect_equal(
  cj(list(1L, "two", TRUE, NULL, FALSE)),
  "[1,\"two\",true,null,false]"
)

# array of objects with their own keys to sort
expect_equal(
  cj(list(list(b = 2L, a = 1L), list(d = 4L, c = 3L))),
  "[{\"a\":1,\"b\":2},{\"c\":3,\"d\":4}]"
)

# I() (AsIs) forces array encoding regardless of underlying length,
# matching jsonlite. Names on the underlying value are dropped.
expect_equal(cj(I(1L)), "[1]")
expect_equal(cj(I("a")), "[\"a\"]")
expect_equal(cj(I(TRUE)), "[true]")
expect_equal(cj(I(c(1L, 2L, 3L))), "[1,2,3]")
expect_equal(cj(I(c("x", "y"))), "[\"x\",\"y\"]")
expect_equal(cj(I(list(a = 1L, b = 2L))), "[1,2]")   # names dropped
expect_equal(cj(I(list())), "[]")
# Per-element validation still applies under I().
expect_error(cj(I(NA)),         pattern = "NA disallowed")
expect_error(cj(I(c(1, NaN))),  pattern = "NA/NaN/Inf disallowed")
expect_error(cj(I(c(1, 1.5))),  pattern = "non-integer")

# =========================================================================
# Objects and key sorting
# =========================================================================

# empty named list — character(0) names means "{}", NULL names means "[]"
expect_equal(cj(structure(list(), names = character(0))), "{}")
expect_equal(cj(list()), "[]")

# basic sorting
expect_equal(cj(list(b = 2L, a = 1L)), "{\"a\":1,\"b\":2}")
expect_equal(
  cj(list(z = "z", a = "a", m = "m")),
  "{\"a\":\"a\",\"m\":\"m\",\"z\":\"z\"}"
)

# mixed-length keys (lexicographic, not by length)
expect_equal(
  cj(list(b = 1L, aaa = 2L, aa = 3L)),
  "{\"aa\":3,\"aaa\":2,\"b\":1}"
)

# Sorting is locale-independent (UTF-8 byte order). Uppercase letters
# (0x41-0x5A) sort BEFORE lowercase (0x61-0x7A) in codepoint order; a
# locale-aware sort can interleave them. Test the failure mode directly.
expect_equal(
  cj(list(b = 1L, A = 2L, a = 3L, B = 4L)),
  "{\"A\":2,\"B\":4,\"a\":3,\"b\":1}"
)

# Special characters in keys, with codepoint ordering verified by hand:
# '!' = 0x21, '.' = 0x2E, '0' = 0x30, '@' = 0x40, '_' = 0x5F, 'a' = 0x61
expect_equal(
  cj(list(`a` = 1L, `_id` = 2L, `@user` = 3L, `0n` = 4L,
          `.dot` = 5L, `!bang` = 6L)),
  paste0("{",
         "\"!bang\":6,",
         "\".dot\":5,",
         "\"0n\":4,",
         "\"@user\":3,",
         "\"_id\":2,",
         "\"a\":1",
         "}")
)

# Unicode keys: 'z' (0x7A) < 'à' (U+00E0) by both codepoint and UTF-8
zlist <- list(1L, 2L)
names(zlist) <- c("z", "à")
expect_equal(cj(zlist), "{\"z\":1,\"à\":2}")

# Deeply nested sorted objects
expect_equal(
  cj(list(
    z = list(b = 1L, a = 2L),
    a = list(y = 3L, x = list(p = list(q = 4L, r = 5L), o = 6L))
  )),
  paste0("{",
         "\"a\":{\"x\":{\"o\":6,\"p\":{\"q\":4,\"r\":5}},\"y\":3},",
         "\"z\":{\"a\":2,\"b\":1}",
         "}")
)

# Object containing arrays of objects, both ends needing sort
expect_equal(
  cj(list(b = list(list(d = 1L, c = 2L)), a = "x")),
  "{\"a\":\"x\",\"b\":[{\"c\":2,\"d\":1}]}"
)

# NULL field in a named list emits "key":null (not dropped)
expect_equal(
  cj(list(a = NULL, b = 1L)),
  "{\"a\":null,\"b\":1}"
)

# NA object keys must be rejected (would silently coerce to the literal
# string "NA" otherwise).
l <- list(1L, 2L)
names(l) <- c("a", NA_character_)
expect_error(cj(l), pattern = "NA in object keys")

# Duplicate object keys must be rejected (RFC 8259 leaves duplicate
# behavior undefined; signing two values under the same key is a recipe
# for ambiguity).
expect_error(cj(list(a = 1L, a = 2L)), pattern = "duplicate object key")
expect_error(
  cj(list(a = 1L, b = 2L, a = 3L)),
  pattern = "duplicate object key"
)
# Duplicate detection runs after UTF-8 normalisation, so byte-identical
# keys with different R encodings still collide.
dup <- list(1L, 2L)
names(dup) <- enc2utf8(c("café", "café"))
expect_error(cj(dup), pattern = "duplicate object key")

# =========================================================================
# Matrix /keys/upload signing payload — full realistic shape
# =========================================================================

dk <- list(
  user_id = "@alice:example.org",
  device_id = "ABCD",
  algorithms = list("m.olm.v1.curve25519-aes-sha2",
                    "m.megolm.v1.aes-sha2"),
  keys = list(
    `ed25519:ABCD` = "EEEE",
    `curve25519:ABCD` = "CCCC"
  )
)
out <- cj(dk)
expect_equal(
  out,
  paste0(
    "{",
    "\"algorithms\":[\"m.olm.v1.curve25519-aes-sha2\",",
    "\"m.megolm.v1.aes-sha2\"],",
    "\"device_id\":\"ABCD\",",
    "\"keys\":{\"curve25519:ABCD\":\"CCCC\",\"ed25519:ABCD\":\"EEEE\"},",
    "\"user_id\":\"@alice:example.org\"",
    "}"
  )
)

# Same payload with a signatures block already attached: signing happens
# on the object minus signatures/unsigned, but cj() itself just emits
# what it's handed — verify the user_id key's '@' sorts before letters.
signed <- dk
signed$signatures <- list(
  `@alice:example.org` = list(`ed25519:ABCD` = "SIGSIGSIG")
)
out2 <- cj(signed)
expect_true(grepl("\"signatures\":\\{", out2))
# '@' (0x40) < lowercase letters, so signatures comes before user_id
expect_true(
  regexpr("\"signatures\"", out2, fixed = TRUE) <
  regexpr("\"user_id\"", out2, fixed = TRUE)
)

# User profile: displayname and avatar.

#' Get a user's profile
#'
#' @param session An "mx_session" object.
#' @param user_id Character. Defaults to the session's own user.
#' @return A list with \code{displayname} and \code{avatar_url} (either
#'   may be absent when unset).
#' @examples
#' \dontrun{
#' mx_profile(s)$displayname
#' }
#' @export
mx_profile <- function(session, user_id = session$user_id) {
    path <- sprintf("/_matrix/client/v3/profile/%s", mx_encode_id(user_id))
    mx_http(session$server, "GET", path, token = session$token)
}

#' Set this user's display name
#'
#' @param session An "mx_session" object.
#' @param displayname Character. The new display name.
#' @return Invisibly TRUE on success.
#' @examples
#' \dontrun{
#' mx_set_displayname(s, "cornelius")
#' }
#' @export
mx_set_displayname <- function(session, displayname) {
    path <- sprintf(
                    "/_matrix/client/v3/profile/%s/displayname",
                    mx_encode_id(session$user_id)
    )
    mx_http(session$server, "PUT", path,
            body = list(displayname = displayname), token = session$token)
    invisible(TRUE)
}

#' Set this user's avatar
#'
#' @param session An "mx_session" object.
#' @param avatar_url Character. An \code{mxc://} URI, typically from
#'   \code{\link{mx_upload}}.
#' @return Invisibly TRUE on success.
#' @examples
#' \dontrun{
#' uri <- mx_upload(s, "avatar.png")
#' mx_set_avatar_url(s, uri)
#' }
#' @export
mx_set_avatar_url <- function(session, avatar_url) {
    path <- sprintf(
                    "/_matrix/client/v3/profile/%s/avatar_url",
                    mx_encode_id(session$user_id)
    )
    mx_http(session$server, "PUT", path,
            body = list(avatar_url = avatar_url), token = session$token)
    invisible(TRUE)
}

# Media upload and download

#' Upload a file to the homeserver media repository
#'
#' @param session An "mx_session" object.
#' @param path Character. Local file path.
#' @param content_type Character or NULL. MIME type; guessed from the
#'   file extension if NULL.
#' @param filename Character or NULL. Filename advertised to the server.
#'
#' @return An "mxc://" URI as a character string.
#' @examples
#' \dontrun{
#' uri <- mx_upload(s, "photo.png")
#' }
#' @export
mx_upload <- function(session, path, content_type = NULL, filename = NULL) {
    if (!file.exists(path)) {
        stop(sprintf("File not found: %s", path), call. = FALSE)
    }
    if (is.null(content_type)) {
        content_type <- mx_guess_mime(path)
    }
    if (is.null(filename)) {
        filename <- basename(path)
    }

    url <- paste0(
                  sub("/$", "", session$server),
                  "/_matrix/media/v3/upload?filename=",
                  utils::URLencode(filename, reserved = TRUE)
    )

    # Stream from disk rather than reading the whole file into RAM:
    # upload mode + a readfunction lets curl pull chunks as it sends,
    # which keeps multi-GB videos out of memory. customrequest keeps
    # the method POST (upload mode alone would PUT).
    con <- file(path, "rb")
    on.exit(close(con), add = TRUE)

    h <- curl::new_handle()
    curl::handle_setopt(
                        h,
                        upload = TRUE,
                        customrequest = "POST",
                        readfunction = function(n) readBin(con, "raw", n),
                        infilesize_large = file.size(path)
    )
    curl::handle_setheaders(
                            h,
                            Authorization = paste("Bearer", session$token),
                            `Content-Type` = content_type,
                            Accept = "application/json"
    )
    resp <- curl::curl_fetch_memory(url, handle = h)
    # A proxy or misconfigured server can answer non-JSON (HTML, plain
    # text); don't let the parse error mask the real failure -- fall
    # through to mx_raise with whatever body came back.
    parsed <- tryCatch(
                       jsonlite::fromJSON(rawToChar(resp$content),
                                          simplifyVector = FALSE),
                       error = function(e) list(raw = rawToChar(resp$content))
    )

    if (resp$status_code >= 400) {
        mx_raise(parsed$errcode %||% "HTTP",
                 parsed$error %||% paste("HTTP", resp$status_code),
                 status = resp$status_code, body = parsed)
    }

    parsed$content_uri
}

#' Download a media file by mxc URI
#'
#' @param session An "mx_session" object.
#' @param mxc_url Character. An "mxc://server/id" URI.
#' @param dest Character. Destination file path.
#'
#' @return The destination path, invisibly.
#' @examples
#' \dontrun{
#' mx_download(s, "mxc://matrix.example/abc123", tempfile())
#' }
#' @export
mx_download <- function(session, mxc_url, dest) {
    m <- regmatches(mxc_url, regexec("^mxc://([^/]+)/(.+)$", mxc_url))[[1]]
    if (length(m) != 3L) {
        stop(sprintf("Not an mxc URI: %s", mxc_url), call. = FALSE)
    }
    server_name <- m[2]
    media_id <- m[3]

    url <- paste0(
                  sub("/$", "", session$server),
                  "/_matrix/client/v1/media/download/",
                  utils::URLencode(server_name, reserved = TRUE), "/",
                  utils::URLencode(media_id, reserved = TRUE)
    )

    h <- curl::new_handle()
    curl::handle_setheaders(h, Authorization = paste("Bearer", session$token))
    curl::curl_download(url, dest, handle = h)
    invisible(dest)
}

#' Guess a MIME type from a file extension
#'
#' The extension table mx.api uses for uploads, exported so callers do
#' not maintain their own. Unknown extensions fall back to
#' \code{"application/octet-stream"}.
#'
#' @param path Character. File path or name.
#' @return Character MIME type.
#' @examples
#' mx_guess_mime("clip.mp4")
#' mx_guess_mime("notes.txt")
#' @export
mx_guess_mime <- function(path) {
    ext <- tolower(tools::file_ext(path))
    table <- c(
               txt = "text/plain", md = "text/markdown", csv = "text/csv",
               json = "application/json", pdf = "application/pdf",
               html = "text/html", xml = "application/xml",
               png = "image/png", jpg = "image/jpeg", jpeg = "image/jpeg",
               gif = "image/gif", webp = "image/webp", svg = "image/svg+xml",
               mp3 = "audio/mpeg", wav = "audio/wav", ogg = "audio/ogg",
               mp4 = "video/mp4", webm = "video/webm", mov = "video/quicktime",
               zip = "application/zip", gz = "application/gzip",
               tar = "application/x-tar"
    )
    hit <- unname(table[ext])
    # A named-vector miss is NA, not NULL, so %||% alone won't catch it.
    if (length(hit) != 1L || is.na(hit)) {
        return("application/octet-stream")
    }
    hit
}


#' Send a media file to a room
#'
#' Uploads \code{path} to the media repository and posts an
#' \code{m.room.message} referencing it. The default \code{info} carries
#' \code{mimetype} and \code{size}; pass richer metadata (width, height,
#' duration) yourself -- mx.api deliberately does not inspect media files.
#'
#' @param session An "mx_session" object.
#' @param room_id Character. The room ID.
#' @param path Character. Path to the file to upload.
#' @param body Character. Message body / filename shown by clients.
#' @param msgtype Character or NULL. One of \code{"m.file"},
#'   \code{"m.image"}, \code{"m.audio"}, \code{"m.video"}. NULL (the
#'   default) derives it from the MIME type, so a .mp4 posts as m.video
#'   without being told.
#' @param content_type Character or NULL. MIME type (guessed from the
#'   extension when NULL).
#' @param info List. Extra fields merged into the \code{info} object.
#' @return The event ID of the sent message.
#' @examples
#' \dontrun{
#' mx_send_media(s, "!abc:example", "clip.mp4", msgtype = "m.video")
#' }
#' @export
mx_send_media <- function(session, room_id, path, body = basename(path),
                          msgtype = NULL, content_type = NULL,
                          info = list()) {
    if (is.null(content_type)) {
        content_type <- mx_guess_mime(path)
    }
    if (is.null(msgtype)) {
        msgtype <- mx_msgtype_for_mime(content_type)
    }
    uri <- mx_upload(session, path, content_type = content_type,
                     filename = basename(path))
    base_info <- list(mimetype = content_type, size = file.size(path))
    if (length(info)) {
        base_info <- utils::modifyList(base_info, info)
    }
    mx_send(session, room_id, body, msgtype = msgtype,
            extra = list(url = uri, info = base_info))
}

#' @rdname mx_send_media
#' @export
mx_send_file <- function(session, room_id, path, body = basename(path),
                         content_type = NULL, info = list()) {
    mx_send_media(session, room_id, path, body = body, msgtype = "m.file",
                  content_type = content_type, info = info)
}

#' @rdname mx_send_media
#' @export
mx_send_image <- function(session, room_id, path, body = basename(path),
                          content_type = NULL, info = list()) {
    mx_send_media(session, room_id, path, body = body, msgtype = "m.image",
                  content_type = content_type, info = info)
}

#' @rdname mx_send_media
#' @export
mx_send_audio <- function(session, room_id, path, body = basename(path),
                          content_type = NULL, info = list()) {
    mx_send_media(session, room_id, path, body = body, msgtype = "m.audio",
                  content_type = content_type, info = info)
}

#' @rdname mx_send_media
#' @export
mx_send_video <- function(session, room_id, path, body = basename(path),
                          content_type = NULL, info = list()) {
    mx_send_media(session, room_id, path, body = body, msgtype = "m.video",
                  content_type = content_type, info = info)
}

# m.image / m.audio / m.video from the MIME family; m.file otherwise.
mx_msgtype_for_mime <- function(content_type) {
    if (startsWith(content_type, "image/")) {
        return("m.image")
    }
    if (startsWith(content_type, "audio/")) {
        return("m.audio")
    }
    if (startsWith(content_type, "video/")) {
        return("m.video")
    }
    "m.file"
}

#' Query the homeserver's media configuration
#'
#' Asks the server for its media limits, chiefly \code{m.upload.size}
#' (maximum upload bytes), so callers can check a file fits before
#' uploading. Tries the v1 endpoint and falls back to the legacy
#' location for older homeservers.
#'
#' @param session An "mx_session" object.
#' @return A list; \code{$`m.upload.size`} is the upload cap in bytes
#'   (may be absent if the server does not advertise one).
#' @examples
#' \dontrun{
#' cap <- mx_media_config(s)$`m.upload.size`
#' file.size("clip.mp4") <= cap
#' }
#' @export
mx_media_config <- function(session) {
    tryCatch(
             mx_http(session$server, "GET",
                     "/_matrix/client/v1/media/config",
                     token = session$token),
             error = function(e) {
        mx_http(session$server, "GET", "/_matrix/media/v3/config",
                token = session$token)
    }
    )
}

library(tinytest)

expect_true(is.function(mx.api::mx_upload))
expect_true(is.function(mx.api::mx_download))

if (at_home() && nzchar(Sys.getenv("MX_TEST_SERVER"))) {
  # live upload/download round-trip goes here
}

# Media message helpers: signatures + msgtype routing.
expect_true(is.function(mx.api::mx_send_media))
expect_equal(
  names(formals(mx.api::mx_send_media)),
  c("session", "room_id", "path", "body", "msgtype", "content_type", "info")
)
for (fn in c("mx_send_file", "mx_send_image", "mx_send_audio",
             "mx_send_video")) {
  expect_true(is.function(getExportedValue("mx.api", fn)))
}

# msgtype auto-detection + public mime guesser + media config.
expect_equal(mx.api:::mx_msgtype_for_mime("video/mp4"), "m.video")
expect_equal(mx.api:::mx_msgtype_for_mime("audio/ogg"), "m.audio")
expect_equal(mx.api:::mx_msgtype_for_mime("image/png"), "m.image")
expect_equal(mx.api:::mx_msgtype_for_mime("application/pdf"), "m.file")
expect_equal(mx.api::mx_guess_mime("clip.MP4"), "video/mp4")
expect_equal(mx.api::mx_guess_mime("unknown.xyz"), "application/octet-stream")
expect_true(is.function(mx.api::mx_media_config))

# Message send, history, and sync

#' Send a message to a room
#'
#' @param session An "mx_session" object.
#' @param room_id Character. The room ID.
#' @param body Character. The message body.
#' @param msgtype Character. Matrix msgtype, default "m.text".
#' @param extra List or NULL. Extra fields merged into the event content
#'   (e.g. formatted body, reply relation).
#'
#' @return The event ID of the sent message.
#' @examples
#' \dontrun{
#' mx_send(s, "!abc:matrix.example", "hello world")
#' }
#' @export
mx_send <- function(session, room_id, body, msgtype = "m.text", extra = NULL) {
    content <- list(msgtype = msgtype, body = body)
    if (length(extra)) {
        content <- utils::modifyList(content, extra)
    }

    path <- sprintf(
                    "/_matrix/client/v3/rooms/%s/send/m.room.message/%s",
                    mx_encode_id(room_id), mx_encode_id(mx_txn_id())
    )
    resp <- mx_http(
                    session$server, "PUT", path,
                    body = content, token = session$token
    )
    resp$event_id
}

#' Send an arbitrary room event
#'
#' Generic counterpart to \code{\link{mx_send}} for event types other than
#' \code{m.room.message}, such as \code{m.room.encrypted}. The content is
#' sent verbatim.
#'
#' @param session An "mx_session" object.
#' @param room_id Character. The room ID.
#' @param event_type Character. The event type, e.g.
#'   \code{"m.room.encrypted"}.
#' @param content List. The event content, sent as-is.
#' @param txn_id Character or NULL. Transaction id (generated if NULL).
#' @return The event ID of the sent event.
#' @examples
#' \dontrun{
#' mx_send_event(s, "!abc:example", "m.room.encrypted", encrypted_content)
#' }
#' @export
mx_send_event <- function(session, room_id, event_type, content,
                          txn_id = NULL) {
    if (is.null(txn_id)) {
        txn_id <- mx_txn_id()
    }
    path <- sprintf(
                    "/_matrix/client/v3/rooms/%s/send/%s/%s",
                    mx_encode_id(room_id), mx_encode_id(event_type),
                    mx_encode_id(txn_id)
    )
    resp <- mx_http(
                    session$server, "PUT", path,
                    body = content, token = session$token
    )
    resp$event_id
}

#' Set a room state event
#'
#' Generic state setter, e.g. to mark a room encrypted by putting an
#' \code{m.room.encryption} event.
#'
#' @param session An "mx_session" object.
#' @param room_id Character. The room ID.
#' @param event_type Character. State event type, e.g.
#'   \code{"m.room.encryption"}.
#' @param content List. The state content, sent as-is.
#' @param state_key Character. State key (default empty string).
#' @return The event ID of the state event.
#' @examples
#' \dontrun{
#' mx_set_state(s, "!abc:example", "m.room.encryption",
#'              list(algorithm = "m.megolm.v1.aes-sha2"))
#' }
#' @export
mx_set_state <- function(session, room_id, event_type, content,
                         state_key = "") {
    path <- sprintf(
                    "/_matrix/client/v3/rooms/%s/state/%s/%s",
                    mx_encode_id(room_id), mx_encode_id(event_type),
                    mx_encode_id(state_key)
    )
    resp <- mx_http(
                    session$server, "PUT", path,
                    body = content, token = session$token
    )
    resp$event_id
}

#' Get a room state event
#'
#' Read-side counterpart of \code{\link{mx_set_state}}, e.g. to check
#' whether a room is encrypted before joining the send path.
#'
#' @param session An "mx_session" object.
#' @param room_id Character. The room ID.
#' @param event_type Character. State event type, e.g.
#'   \code{"m.room.encryption"}.
#' @param state_key Character. State key (default empty string).
#' @return The state event content as a list, or NULL when the state
#'   event is not set.
#' @examples
#' \dontrun{
#' enc <- mx_get_state(s, "!abc:example", "m.room.encryption")
#' is.null(enc)   # FALSE in an encrypted room
#' }
#' @export
mx_get_state <- function(session, room_id, event_type, state_key = "") {
    path <- sprintf(
                    "/_matrix/client/v3/rooms/%s/state/%s/%s",
                    mx_encode_id(room_id), mx_encode_id(event_type),
                    mx_encode_id(state_key)
    )
    tryCatch(
             mx_http(session$server, "GET", path, token = session$token),
             mx_error_M_NOT_FOUND = function(e) NULL
    )
}

#' Fetch historical messages from a room
#'
#' Thin wrapper over the /rooms/{id}/messages endpoint.
#'
#' @param session An "mx_session" object.
#' @param room_id Character. The room ID.
#' @param from Character or NULL. Pagination token; NULL starts at the
#'   most recent message.
#' @param dir Character. "b" (backwards, default) or "f" (forwards).
#' @param limit Integer. Maximum events to return.
#'
#' @return A list with fields chunk, start, end.
#' @examples
#' \dontrun{
#' mx_messages(s, "!abc:matrix.example", limit = 20L)
#' }
#' @export
mx_messages <- function(session, room_id, from = NULL, dir = "b", limit = 50L) {
    query <- list(dir = dir, limit = as.integer(limit))
    if (!is.null(from)) {
        query$from <- from
    }

    path <- sprintf(
                    "/_matrix/client/v3/rooms/%s/messages",
                    mx_encode_id(room_id)
    )
    mx_http(session$server, "GET", path, query = query, token = session$token)
}

#' Send a read receipt for a room event
#'
#' Public receipt (\code{m.read}) advances the "seen" marker in other
#' clients; private receipt (\code{m.read.private}) only advances the
#' bot's own view. Defaults to public so user clients show
#' "seen by @bot".
#'
#' @param session An "mx_session" object.
#' @param room_id Character. The room ID.
#' @param event_id Character. The event to mark as read.
#' @param receipt_type Character. "m.read" (default) or "m.read.private".
#'
#' @return Invisible NULL.
#' @examples
#' \dontrun{
#' mx_read_receipt(s, "!abc:matrix.example", "$eventid")
#' }
#' @export
mx_read_receipt <- function(session, room_id, event_id,
                            receipt_type = c("m.read", "m.read.private")) {
    receipt_type <- match.arg(receipt_type)
    path <- sprintf(
                    "/_matrix/client/v3/rooms/%s/receipt/%s/%s",
                    mx_encode_id(room_id),
                    mx_encode_id(receipt_type),
                    mx_encode_id(event_id)
    )
    mx_http(
            session$server, "POST", path,
            body = mx_empty_body(), token = session$token
    )
    invisible(NULL)
}

#' Send a reaction (annotation) to a room event
#'
#' Posts an m.reaction event tying \code{key} (usually a thumbs-up or
#' other emoji) to \code{event_id}. Matrix reactions are plain events
#' under the hood; they relate to the target via \code{m.annotation}.
#'
#' @param session An "mx_session" object.
#' @param room_id Character. The room ID.
#' @param event_id Character. The event being reacted to.
#' @param key Character. The reaction key (usually an emoji).
#'
#' @return The event ID of the sent reaction.
#' @examples
#' \dontrun{
#' mx_react(s, "!abc:matrix.example", "$eventid", "thumbs-up")
#' }
#' @export
mx_react <- function(session, room_id, event_id, key) {
    content <- list(
                    `m.relates_to` = list(
            rel_type = "m.annotation",
            event_id = event_id,
            key = key
        )
    )
    path <- sprintf(
                    "/_matrix/client/v3/rooms/%s/send/m.reaction/%s",
                    mx_encode_id(room_id), mx_encode_id(mx_txn_id())
    )
    resp <- mx_http(
                    session$server, "PUT", path,
                    body = content, token = session$token
    )
    resp$event_id
}

#' One-shot sync against the homeserver
#'
#' Calls /sync once and returns immediately. For streaming behaviour,
#' the caller writes its own loop, passing the previous batch's
#' next_batch token as \code{since}.
#'
#' @param session An "mx_session" object.
#' @param since Character or NULL. Sync token from a prior sync.
#' @param timeout Integer. Long-poll timeout in milliseconds (0 returns
#'   immediately).
#' @param filter Character or NULL. Filter ID or inline JSON filter.
#'
#' @return The parsed sync response, including next_batch.
#' @examples
#' \dontrun{
#' batch <- mx_sync(s)
#' next_batch <- batch$next_batch
#' }
#' @export
mx_sync <- function(session, since = NULL, timeout = 0L, filter = NULL) {
    query <- list(timeout = as.integer(timeout))
    if (!is.null(since)) {
        query$since <- since
    }
    if (!is.null(filter)) {
        query$filter <- filter
    }

    mx_http(
            session$server, "GET", "/_matrix/client/v3/sync",
            query = query, token = session$token
    )
}


#' Redact an event
#'
#' Removes the content of a message, reaction, or other event. This is
#' how Matrix deletes things.
#'
#' @param session An "mx_session" object.
#' @param room_id Character. The room ID.
#' @param event_id Character. The event to redact.
#' @param reason Character or NULL. Optional human-readable reason.
#' @param txn_id Character or NULL. Transaction id (generated if NULL).
#' @return The event ID of the redaction event.
#' @examples
#' \dontrun{
#' mx_redact(s, "!abc:example", "$someevent", reason = "typo")
#' }
#' @export
mx_redact <- function(session, room_id, event_id, reason = NULL,
                      txn_id = NULL) {
    if (is.null(txn_id)) {
        txn_id <- mx_txn_id()
    }
    body <- if (is.null(reason)) {
        mx_empty_body()
    } else {
        list(reason = reason)
    }
    path <- sprintf(
                    "/_matrix/client/v3/rooms/%s/redact/%s/%s",
                    mx_encode_id(room_id), mx_encode_id(event_id),
                    mx_encode_id(txn_id)
    )
    resp <- mx_http(session$server, "PUT", path, body = body,
                    token = session$token)
    resp$event_id
}

#' Send a typing notification
#'
#' Shows (or clears) the session user's typing indicator in a room.
#' Useful bot polish while a slow reply is being generated.
#'
#' @param session An "mx_session" object.
#' @param room_id Character. The room ID.
#' @param typing Logical. TRUE to show typing, FALSE to clear it.
#' @param timeout Integer. How long the indicator lasts, in
#'   milliseconds (ignored when \code{typing = FALSE}).
#' @return Invisibly TRUE on success.
#' @examples
#' \dontrun{
#' mx_typing(s, "!abc:example", TRUE)
#' # ... generate the reply ...
#' mx_typing(s, "!abc:example", FALSE)
#' }
#' @export
mx_typing <- function(session, room_id, typing = TRUE, timeout = 30000L) {
    body <- list(typing = isTRUE(typing))
    if (isTRUE(typing)) {
        body$timeout <- as.integer(timeout)
    }
    path <- sprintf(
                    "/_matrix/client/v3/rooms/%s/typing/%s",
                    mx_encode_id(room_id), mx_encode_id(session$user_id)
    )
    mx_http(session$server, "PUT", path, body = body,
            token = session$token)
    invisible(TRUE)
}

library(tinytest)

expect_true(is.function(mx.api::mx_send))
expect_true(is.function(mx.api::mx_messages))
expect_true(is.function(mx.api::mx_sync))
expect_true(is.function(mx.api::mx_react))
expect_true(is.function(mx.api::mx_read_receipt))

if (at_home() && nzchar(Sys.getenv("MX_TEST_SERVER")) &&
    nzchar(Sys.getenv("MX_TEST_ROOM"))) {
  s <- mx.api::mx_login(
    Sys.getenv("MX_TEST_SERVER"),
    Sys.getenv("MX_TEST_USER"),
    Sys.getenv("MX_TEST_PASS")
  )
  room <- Sys.getenv("MX_TEST_ROOM")

  eid <- mx.api::mx_send(s, room, "mx.api test")
  expect_true(is.character(eid) && nchar(eid) > 0)

  sync <- mx.api::mx_sync(s, timeout = 0)
  expect_true(!is.null(sync$next_batch))

  mx.api::mx_logout(s)
}

# Canonical JSON per the Matrix spec — intentionally hand-rolled,
# not a wrapper around jsonlite.
#
# Matrix request signing needs byte-stable canonical JSON: every
# implementation must produce the exact same bytes for the same input
# because those bytes are what gets fed into ed25519. A general-purpose
# JSON serializer's job is to produce *valid* JSON; that's a strictly
# weaker contract. Key ordering, number formatting (no scientific
# notation, integer-vs-double behavior), NaN / Inf / NA handling, and
# whether non-ASCII gets emitted as \uXXXX escapes or as raw UTF-8
# bytes are all free choices for a general serializer that the
# canonical-JSON spec pins down.
#
# This file keeps every one of those rules visible and directly
# testable in roughly 80 lines. The alternative — relying on
# "jsonlite's defaults happen to match the spec (today)" — moves the
# correctness of every signature into another package's release notes.
#
# Spec: https://spec.matrix.org/v1.10/appendices/#canonical-json
#
# - Object keys sorted by UTF-8 byte sequence (radix sort, locale-free)
# - No insignificant whitespace
# - Non-ASCII preserved as raw UTF-8 bytes (no \uXXXX escapes)
# - Numbers: integers only in [-(2^53)+1, (2^53)-1]; no decimal places,
#   no exponents; floats, NaN, Inf, NA all rejected (spec: "Float
#   values are not permitted by this encoding")
# - Object keys: must not be NA or duplicated
# - Control chars 0x01-0x1F escaped (\b \f \n \r \t as named, rest \u)

#' Encode a value as Matrix canonical JSON
#'
#' Produces the canonical JSON byte sequence the Matrix specification
#' requires for signed objects: object keys sorted by UTF-8 byte
#' sequence, no insignificant whitespace, raw UTF-8 for non-ASCII
#' strings, integers only (no floats, no exponents, no decimal places,
#' within \code{[-(2^53)+1, (2^53)-1]}), and rejection of NaN, Inf,
#' NA values, and NA or duplicate object keys. The output is the exact
#' byte sequence to feed into an ed25519 signer.
#'
#' R named lists become JSON objects; unnamed lists and length > 1
#' atomic vectors become arrays. Length-1 atomics become scalars. To
#' force a length-1 value to encode as an array, wrap it in a
#' single-element \code{list(...)} or in \code{I()} (AsIs values are
#' always encoded as arrays, names dropped, mirroring jsonlite).
#'
#' @param x An R value: NULL, atomic vector, or list.
#'
#' @return A length-1 UTF-8 character string. The result is the exact
#'   byte sequence to write to disk or feed to an ed25519 signer;
#'   non-ASCII content is preserved as raw UTF-8 bytes
#'   (jsonlite-style \\uXXXX escaping is not used).
#'
#' @examples
#' mx_canonical_json(list(b = 2, a = 1))
#' # "{\"a\":1,\"b\":2}"
#'
#' mx_canonical_json(list(key = "abc"))
#' # "{\"key\":\"abc\"}"
#'
#' @export
mx_canonical_json <- function(x) {
    mx_cj_emit(x)
}

mx_cj_emit <- function(x) {
    if (is.null(x)) {
        return("null")
    }
    if (inherits(x, "AsIs")) {
        # I() forces array encoding regardless of length, matching the
        # jsonlite convention. Coerce to an unnamed list and recurse.
        x <- unclass(x)
        if (is.atomic(x)) {
            x <- as.list(x)
        } else if (is.list(x)) {
            x <- unname(x)
        }
        return(mx_cj_emit(x))
    }
    if (is.list(x)) {
        nm <- names(x)
        if (length(x) == 0L) {
            return(if (is.null(nm)) "[]" else "{}")
        }
        if (is.null(nm)) {
            return(paste0(
                          "[",
                          paste(vapply(x, mx_cj_emit, character(1)), collapse = ","),
                          "]"
                ))
        }
        if (any(is.na(nm))) {
            stop("mx_canonical_json: NA in object keys disallowed",
                 call. = FALSE)
        }
        nm_utf8 <- enc2utf8(nm)
        if (anyDuplicated(nm_utf8)) {
            dup <- nm_utf8[duplicated(nm_utf8)][[1L]]
            stop(sprintf(
                         "mx_canonical_json: duplicate object key '%s'", dup
                ), call. = FALSE)
        }
        ord <- order(nm_utf8, method = "radix")
        parts <- vapply(ord, function(i) {
            paste0(mx_cj_string(nm_utf8[[i]]), ":", mx_cj_emit(x[[i]]))
        }, character(1))
        return(paste0("{", paste(parts, collapse = ","), "}"))
    }
    if (is.logical(x)) {
        if (any(is.na(x))) {
            stop("mx_canonical_json: NA disallowed", call. = FALSE)
        }
        if (length(x) == 1L) {
            return(if (isTRUE(x)) "true" else "false")
        }
        return(paste0(
                      "[",
                      paste(ifelse(x, "true", "false"), collapse = ","),
                      "]"
            ))
    }
    if (is.character(x)) {
        if (any(is.na(x))) {
            stop("mx_canonical_json: NA disallowed", call. = FALSE)
        }
        if (length(x) == 1L) {
            return(mx_cj_string(x))
        }
        return(paste0(
                      "[",
                      paste(vapply(x, mx_cj_string, character(1)), collapse = ","),
                      "]"
            ))
    }
    if (is.numeric(x)) {
        if (any(is.na(x)) || any(!is.finite(x))) {
            stop("mx_canonical_json: NA/NaN/Inf disallowed", call. = FALSE)
        }
        if (length(x) == 1L) {
            return(mx_cj_number(x))
        }
        return(paste0(
                      "[",
                      paste(vapply(x, mx_cj_number, character(1)), collapse = ","),
                      "]"
            ))
    }
    stop(sprintf("mx_canonical_json: unsupported type '%s'", typeof(x)),
         call. = FALSE)
}

mx_cj_number <- function(x) {
    # Matrix canonical JSON allows only integers in
    # [-(2^53)+1, (2^53)-1] (spec: "Float values are not permitted by
    # this encoding").
    if (x != trunc(x)) {
        stop(sprintf(
                     "mx_canonical_json: non-integer number %s disallowed", x
            ), call. = FALSE)
    }
    if (abs(x) > 2 ^ 53 - 1) {
        stop(sprintf(
                     "mx_canonical_json: integer %s out of [-2^53+1, 2^53-1]", x
            ), call. = FALSE)
    }
    # R normalises negative zero to "0", which is what the spec
    # requires ("-0 MUST NOT appear" in output).
    format(x, scientific = FALSE, trim = TRUE, drop0trailing = TRUE)
}

mx_cj_string <- function(s) {
    s <- enc2utf8(as.character(s))
    s <- gsub("\\", "\\\\", s, fixed = TRUE)
    s <- gsub("\"", "\\\"", s, fixed = TRUE)
    s <- gsub("\b", "\\b", s, fixed = TRUE)
    s <- gsub("\f", "\\f", s, fixed = TRUE)
    s <- gsub("\n", "\\n", s, fixed = TRUE)
    s <- gsub("\r", "\\r", s, fixed = TRUE)
    s <- gsub("\t", "\\t", s, fixed = TRUE)
    # Any remaining 0x01-0x1F control chars need \u escaping. NUL (0x00)
    # cannot appear in an R string, so it's not iterated.
    for (cp in c(1:7, 11L, 14:31)) {
        ch <- intToUtf8(cp)
        if (grepl(ch, s, fixed = TRUE)) {
            s <- gsub(ch, sprintf("\\u%04x", cp), s, fixed = TRUE)
        }
    }
    paste0("\"", s, "\"")
}

# Device management.

#' List this account's devices
#'
#' @param session An "mx_session" object.
#' @return A list of devices, each with \code{device_id},
#'   \code{display_name}, \code{last_seen_ip}, and \code{last_seen_ts}.
#' @examples
#' \dontrun{
#' vapply(mx_devices(s), function(d) d$device_id, character(1))
#' }
#' @export
mx_devices <- function(session) {
    resp <- mx_http(session$server, "GET", "/_matrix/client/v3/devices",
                    token = session$token)
    resp$devices %||% list()
}

#' Delete a device
#'
#' Removes a device and invalidates its access token. Most homeservers
#' protect this with user-interactive authentication: the first call
#' fails with M_FORBIDDEN or a 401 carrying a \code{flows} object, and
#' the caller retries with a completed \code{auth} payload, e.g.
#' \code{list(type = "m.login.password", identifier = list(type =
#' "m.id.user", user = "bot"), password = "...", session = "<from the
#' 401>")}. mx.api deliberately does not automate that exchange.
#'
#' @param session An "mx_session" object.
#' @param device_id Character. The device to delete.
#' @param auth List or NULL. Completed user-interactive auth payload.
#' @return Invisibly TRUE on success.
#' @examples
#' \dontrun{
#' mx_delete_device(s, "OLDDEVICE", auth = list(
#'     type = "m.login.password",
#'     identifier = list(type = "m.id.user", user = "bot"),
#'     password = "secret"
#' ))
#' }
#' @export
mx_delete_device <- function(session, device_id, auth = NULL) {
    body <- if (is.null(auth)) {
        mx_empty_body()
    } else {
        list(auth = auth)
    }
    path <- sprintf(
                    "/_matrix/client/v3/devices/%s",
                    mx_encode_id(device_id)
    )
    mx_http(session$server, "DELETE", path, body = body,
            token = session$token)
    invisible(TRUE)
}

# Account data: per-user key-value storage on the homeserver.

#' Get account data
#'
#' Reads a global account-data event for a user, e.g. \code{"m.direct"}
#' (the DM room map) or any custom namespaced type.
#'
#' @param session An "mx_session" object.
#' @param type Character. Event type, e.g. \code{"m.direct"}.
#' @param user_id Character. Defaults to the session's own user.
#' @return The account-data content as a list, or NULL when the type has
#'   never been set.
#' @examples
#' \dontrun{
#' direct <- mx_get_account_data(s, "m.direct")
#' }
#' @export
mx_get_account_data <- function(session, type, user_id = session$user_id) {
    path <- sprintf(
                    "/_matrix/client/v3/user/%s/account_data/%s",
                    mx_encode_id(user_id), mx_encode_id(type)
    )
    tryCatch(
             mx_http(session$server, "GET", path, token = session$token),
             mx_error_M_NOT_FOUND = function(e) NULL
    )
}

#' Set account data
#'
#' Writes a global account-data event for a user. The content replaces
#' whatever was stored under \code{type}.
#'
#' @param session An "mx_session" object.
#' @param type Character. Event type, e.g. \code{"m.direct"}.
#' @param content List. The content, sent as-is.
#' @param user_id Character. Defaults to the session's own user.
#' @return Invisibly TRUE on success.
#' @examples
#' \dontrun{
#' mx_set_account_data(s, "ai.cornball.notes", list(theme = "dark"))
#' }
#' @export
mx_set_account_data <- function(session, type, content,
                                user_id = session$user_id) {
    path <- sprintf(
                    "/_matrix/client/v3/user/%s/account_data/%s",
                    mx_encode_id(user_id), mx_encode_id(type)
    )
    mx_http(session$server, "PUT", path, body = content,
            token = session$token)
    invisible(TRUE)
}

#' mx.api: Minimal Matrix Client-Server API
#'
#' Base-R bindings for the Matrix Client-Server HTTP API, suitable for
#' talking to a Synapse homeserver. Two dependencies: curl and jsonlite.
#' End-to-end encryption is out of scope; use unencrypted rooms or a
#' separate crypto package.
#'
#' @name mx.api-package
#' @aliases mx.api
"_PACKAGE"

# Internal HTTP helper. Not exported.

mx_http <- function(base_url, method, path, body = NULL, query = NULL,
                    token = NULL) {
    url <- paste0(sub("/$", "", base_url), path)
    if (length(query)) {
        pairs <- mapply(
                        function(k, v) paste0(utils::URLencode(k, reserved = TRUE), "=",
                utils::URLencode(as.character(v), reserved = TRUE)),
                        names(query), query, SIMPLIFY = TRUE, USE.NAMES = FALSE
        )
        url <- paste0(url, "?", paste(pairs, collapse = "&"))
    }

    h <- curl::new_handle()
    curl::handle_setopt(h, customrequest = method)
    headers <- c(Accept = "application/json")
    if (!is.null(token)) {
        headers <- c(headers, Authorization = paste("Bearer", token))
    }
    if (!is.null(body)) {
        payload <- jsonlite::toJSON(body, auto_unbox = TRUE, null = "null")
        curl::handle_setopt(h, postfields = payload)
        headers <- c(headers, `Content-Type` = "application/json")
    }
    curl::handle_setheaders(h, .list = as.list(headers))

    resp <- curl::curl_fetch_memory(url, handle = h)
    raw <- rawToChar(resp$content)
    parsed <- if (nzchar(raw)) {
        tryCatch(
                 jsonlite::fromJSON(raw, simplifyVector = FALSE),
                 error = function(e) list(raw = raw)
        )
    } else {
        list()
    }

    if (resp$status_code >= 400) {
        mx_raise(parsed$errcode %||% "HTTP",
                 parsed$error %||% paste("HTTP", resp$status_code),
                 status = resp$status_code, body = parsed)
    }

    parsed
}

# Raise a classed Matrix error. Code can catch specific failures by
# class -- e.g. tryCatch(..., mx_error_M_NOT_FOUND = function(e) NULL)
# or test inherits(e, "mx_error_M_UNKNOWN_TOKEN") to re-login -- instead
# of grepl()ing the message. The message text keeps the historical
# "Matrix error [CODE]: msg" shape, and $errcode, $status, and $body
# carry the structured details.
mx_raise <- function(errcode, msg, status = NULL, body = NULL) {
    cond <- structure(
                      class = c(paste0("mx_error_", errcode), "mx_error",
                                "error", "condition"),
                      list(
                           message = sprintf("Matrix error [%s]: %s",
                                             errcode, msg),
                           call = NULL,
                           errcode = errcode,
                           status = status,
                           body = body
                      )
    )
    stop(cond)
}

`%||%` <- function(a, b) if (is.null(a)) b else a

mx_txn_id <- function() {
    paste0("mx-", as.integer(Sys.time()), "-",
           sample.int(.Machine$integer.max, 1))
}

mx_encode_id <- function(x) utils::URLencode(x, reserved = TRUE)

mx_empty_body <- function() stats::setNames(list(), character())

library(tinytest)

# mx_raise() produces classed conditions with structured fields and the
# historical message shape.
e <- tryCatch(mx.api:::mx_raise("M_NOT_FOUND", "Event not found.",
                                status = 404L),
              error = function(e) e)
expect_true(inherits(e, "mx_error"))
expect_true(inherits(e, "mx_error_M_NOT_FOUND"))
expect_equal(e$errcode, "M_NOT_FOUND")
expect_equal(e$status, 404L)
expect_equal(conditionMessage(e),
             "Matrix error [M_NOT_FOUND]: Event not found.")

# class-targeted handlers fire
got <- tryCatch(mx.api:::mx_raise("M_UNKNOWN_TOKEN", "expired"),
                mx_error_M_UNKNOWN_TOKEN = function(e) "relogin")
expect_equal(got, "relogin")

# generic mx_error handler catches any code
got2 <- tryCatch(mx.api:::mx_raise("M_LIMIT_EXCEEDED", "slow down"),
                 mx_error = function(e) e$errcode)
expect_equal(got2, "M_LIMIT_EXCEEDED")

# Device-key and to-device transport endpoints.
#
# These four endpoints carry the raw JSON payloads that an end-to-end
# encryption layer (e.g. mx.crypto) needs to coordinate Olm sessions.
# mx.api itself does no cryptography: signing, OTK generation, and
# Olm / Megolm orchestration live in the caller. The bodies passed in
# must already be signed where the Matrix spec requires it.

#' Upload device identity and one-time keys
#'
#' POST \code{/_matrix/client/v3/keys/upload}. The \code{device_keys}
#' and \code{one_time_keys} arguments must be fully formed and signed
#' per the Matrix specification; mx.api will not canonicalise or sign
#' them. Use \code{\link{mx_canonical_json}} to produce the byte
#' sequence to sign.
#'
#' @param session An \code{mx_session}.
#' @param device_keys Named list. The \code{device_keys} object as
#'   defined in the Matrix spec, including \code{user_id},
#'   \code{device_id}, \code{algorithms}, \code{keys}, and the
#'   \code{signatures} block produced by the caller's signer. Pass
#'   NULL to upload only one-time keys.
#' @param one_time_keys Named list or NULL. Map from
#'   \code{"<algorithm>:<key_id>"} (e.g. \code{"signed_curve25519:AAAA"})
#'   to the signed key object. Pass NULL or an empty list to skip.
#' @param fallback_keys Named list or NULL. Same shape as
#'   \code{one_time_keys}; used to advertise a fallback key when the
#'   OTK pool is exhausted.
#'
#' @return The parsed homeserver response, including
#'   \code{one_time_key_counts}.
#'
#' @examples
#' \dontrun{
#' mx_keys_upload(s, device_keys = signed_dk, one_time_keys = signed_otks)
#' }
#' @export
mx_keys_upload <- function(session, device_keys = NULL, one_time_keys = NULL,
                           fallback_keys = NULL) {
    body <- list()
    if (!is.null(device_keys)) {
        body$device_keys <- device_keys
    }
    if (length(one_time_keys)) {
        body$one_time_keys <- one_time_keys
    }
    if (length(fallback_keys)) {
        body$fallback_keys <- fallback_keys
    }
    if (length(body) == 0L) {
        stop("mx_keys_upload: nothing to upload", call. = FALSE)
    }
    mx_http(session$server, "POST", "/_matrix/client/v3/keys/upload",
            body = body, token = session$token)
}

#' Query device keys for one or more users
#'
#' POST \code{/_matrix/client/v3/keys/query}. Each entry in
#' \code{device_keys} maps a Matrix user id to a character vector of
#' device ids to request. An empty character vector requests all of
#' the user's devices.
#'
#' @param session An \code{mx_session}.
#' @param device_keys Named list. Names are Matrix user ids
#'   (e.g. \code{"@@alice:example.org"}); values are character vectors
#'   of device ids, or \code{character(0)} for "all devices".
#' @param timeout Integer milliseconds. Time the server should wait
#'   for remote homeservers before returning a partial result.
#'
#' @return Parsed response with a \code{device_keys} map of
#'   \code{user_id -> device_id -> device_keys_object}.
#'
#' @examples
#' \dontrun{
#' mx_keys_query(s, list("@alice:example.org" = character()))
#' }
#' @export
mx_keys_query <- function(session, device_keys, timeout = 10000L) {
    if (!is.list(device_keys) || is.null(names(device_keys))) {
        stop("mx_keys_query: 'device_keys' must be a named list", call. = FALSE)
    }
    dk <- lapply(device_keys, function(v) {
        if (is.null(v)) {
            return(list())
        }
        as.list(as.character(v))
    })
    body <- list(device_keys = dk, timeout = as.integer(timeout))
    mx_http(session$server, "POST", "/_matrix/client/v3/keys/query",
            body = body, token = session$token)
}

#' Claim one-time keys for an Olm handshake
#'
#' POST \code{/_matrix/client/v3/keys/claim}. The
#' \code{one_time_keys} argument selects which algorithm to claim for
#' each \code{(user_id, device_id)} pair.
#'
#' @param session An \code{mx_session}.
#' @param one_time_keys Named list. Names are user ids; values are
#'   named lists mapping device ids to the desired algorithm
#'   (typically \code{"signed_curve25519"}).
#' @param timeout Integer milliseconds. Server-side timeout when
#'   talking to remote homeservers.
#'
#' @return Parsed response with the claimed keys keyed by
#'   \code{user_id -> device_id -> "<algorithm>:<key_id>" -> key_object}.
#'
#' @examples
#' \dontrun{
#' mx_keys_claim(s, list(
#'   "@alice:example.org" = list(ABCD1234 = "signed_curve25519")
#' ))
#' }
#' @export
mx_keys_claim <- function(session, one_time_keys, timeout = 10000L) {
    if (!is.list(one_time_keys) || is.null(names(one_time_keys))) {
        stop("mx_keys_claim: 'one_time_keys' must be a named list",
             call. = FALSE)
    }
    body <- list(one_time_keys = one_time_keys, timeout = as.integer(timeout))
    mx_http(session$server, "POST", "/_matrix/client/v3/keys/claim",
            body = body, token = session$token)
}

#' Send a to-device event
#'
#' PUT \code{/_matrix/client/v3/sendToDevice/{eventType}/{txnId}}.
#' Used to ship encrypted Olm payloads (e.g. m.room_key carriers
#' wrapped as \code{m.room.encrypted}) to specific
#' \code{(user_id, device_id)} targets.
#'
#' @param session An \code{mx_session}.
#' @param event_type Character. The to-device event type, e.g.
#'   \code{"m.room.encrypted"}.
#' @param messages Named list. Outer names are user ids; values are
#'   named lists mapping device id (or the wildcard \code{"*"}) to
#'   the event content.
#' @param txn_id Character or NULL. Idempotency key; auto-generated
#'   when NULL.
#'
#' @return Invisible NULL (the server returns an empty body on success).
#'
#' @examples
#' \dontrun{
#' mx_send_to_device(s, "m.room.encrypted", list(
#'   "@bob:example.org" = list(BBBB = encrypted_content)
#' ))
#' }
#' @export
mx_send_to_device <- function(session, event_type, messages, txn_id = NULL) {
    if (!is.list(messages) || is.null(names(messages))) {
        stop("mx_send_to_device: 'messages' must be a named list",
             call. = FALSE)
    }
    if (is.null(txn_id)) {
        txn_id <- mx_txn_id()
    }
    path <- sprintf(
                    "/_matrix/client/v3/sendToDevice/%s/%s",
                    mx_encode_id(as.character(event_type)),
                    mx_encode_id(as.character(txn_id))
    )
    mx_http(session$server, "PUT", path,
            body = list(messages = messages), token = session$token)
    invisible(NULL)
}

library(tinytest)

# Smoke: functions exist with the expected signatures.
expect_true(is.function(mx.api::mx_keys_upload))
expect_true(is.function(mx.api::mx_keys_query))
expect_true(is.function(mx.api::mx_keys_claim))
expect_true(is.function(mx.api::mx_send_to_device))

expect_equal(
  names(formals(mx.api::mx_keys_upload)),
  c("session", "device_keys", "one_time_keys", "fallback_keys")
)
expect_equal(
  names(formals(mx.api::mx_keys_query)),
  c("session", "device_keys", "timeout")
)
expect_equal(
  names(formals(mx.api::mx_keys_claim)),
  c("session", "one_time_keys", "timeout")
)
expect_equal(
  names(formals(mx.api::mx_send_to_device)),
  c("session", "event_type", "messages", "txn_id")
)

# Input validation (no network)
fake <- mx.api::mx_session(
  server = "https://example",
  token = "tok",
  user_id = "@u:example",
  device_id = "DEV"
)
expect_error(mx.api::mx_keys_upload(fake))  # nothing to upload
expect_error(mx.api::mx_keys_query(fake, device_keys = "not a list"))
expect_error(mx.api::mx_keys_claim(fake, one_time_keys = list()))  # unnamed
expect_error(mx.api::mx_send_to_device(fake, "m.room.encrypted",
                                       messages = list()))  # unnamed

# Live round-trip gated on at_home() + env vars. Reuses mx.crypto if
# present so the upload payload is real (signed) — falls back to a
# format-only smoke test otherwise.
if (at_home() && nzchar(Sys.getenv("MX_TEST_SERVER"))) {
  s <- mx.api::mx_login(
    Sys.getenv("MX_TEST_SERVER"),
    Sys.getenv("MX_TEST_USER"),
    Sys.getenv("MX_TEST_PASS")
  )
  on.exit(try(mx.api::mx_logout(s), silent = TRUE), add = TRUE)

  if (requireNamespace("mx.crypto", quietly = TRUE)) {
    acct <- mx.crypto::mxc_account_new()
    ids <- mx.crypto::mxc_account_identity_keys(acct)
    did <- s$device_id

    dk_unsigned <- list(
      user_id = s$user_id,
      device_id = did,
      algorithms = list("m.olm.v1.curve25519-aes-sha2",
                        "m.megolm.v1.aes-sha2"),
      keys = setNames(
        list(ids$curve25519, ids$ed25519),
        c(paste0("curve25519:", did), paste0("ed25519:", did))
      )
    )
    sig <- mx.crypto::mxc_account_sign(
      acct, mx.api::mx_canonical_json(dk_unsigned)
    )
    dk <- dk_unsigned
    dk$signatures <- setNames(
      list(setNames(list(sig), paste0("ed25519:", did))),
      s$user_id
    )

    mx.crypto::mxc_account_generate_one_time_keys(acct, 3L)
    otks <- mx.crypto::mxc_account_one_time_keys(acct)
    signed_otks <- list()
    for (kid in names(otks)) {
      u <- list(key = otks[[kid]])
      ss <- mx.crypto::mxc_account_sign(acct, mx.api::mx_canonical_json(u))
      signed <- u
      signed$signatures <- setNames(
        list(setNames(list(ss), paste0("ed25519:", did))),
        s$user_id
      )
      signed_otks[[paste0("signed_curve25519:", kid)]] <- signed
    }

    resp <- mx.api::mx_keys_upload(
      s, device_keys = dk, one_time_keys = signed_otks
    )
    expect_true(!is.null(resp$one_time_key_counts))
    expect_true(resp$one_time_key_counts$signed_curve25519 >= 3L)

    q <- mx.api::mx_keys_query(s, setNames(list(did), s$user_id))
    expect_true(!is.null(q$device_keys[[s$user_id]][[did]]))

    # Self-claim: works on most homeservers and avoids needing a second user.
    cl <- mx.api::mx_keys_claim(
      s, setNames(list(setNames(list("signed_curve25519"), did)),
                  s$user_id)
    )
    expect_true(!is.null(cl$one_time_keys[[s$user_id]][[did]]))

    # send_to_device to self: harmless empty content.
    mx.api::mx_send_to_device(
      s, "m.test.noop",
      setNames(list(setNames(list(list(ping = "ok")), did)), s$user_id)
    )
  }
}

# Auth and session handling

#' Register a new account on a Matrix homeserver
#'
#' Creates a new user via POST \code{/_matrix/client/v3/register} using
#' the \code{m.login.dummy} auth flow. Most homeservers only accept this
#' when open registration is enabled (or a registration token is
#' supplied). On success returns a ready-to-use \code{mx_session} —
#' registration also logs the new user in.
#'
#' @param server Character. Homeserver base URL.
#' @param username Character. Desired localpart (e.g. "alice").
#' @param password Character. Account password.
#' @param device_id Character or NULL. Device id to assign; a server-
#'   generated one is used if NULL.
#' @param initial_device_display_name Character or NULL. Human-readable
#'   label for the device.
#' @param inhibit_login Logical. When TRUE, the server creates the
#'   account but does not return a session; the call returns a list
#'   with the new \code{user_id} instead of an \code{mx_session}.
#'
#' @return An \code{mx_session} object on login, or a list with
#'   \code{user_id} when \code{inhibit_login = TRUE}.
#' @examples
#' \dontrun{
#' s <- mx_register("https://matrix.example", "alice", "hunter2")
#' }
#' @export
mx_register <- function(server, username, password, device_id = NULL,
                        initial_device_display_name = NULL,
                        inhibit_login = FALSE) {
    body <- list(
                 username = username,
                 password = password,
                 auth = list(type = "m.login.dummy"),
                 inhibit_login = isTRUE(inhibit_login)
    )
    if (!is.null(device_id)) {
        body$device_id <- device_id
    }
    if (!is.null(initial_device_display_name)) {
        body$initial_device_display_name <- initial_device_display_name
    }

    resp <- mx_http(server, "POST", "/_matrix/client/v3/register", body = body)

    if (isTRUE(inhibit_login)) {
        return(list(user_id = resp$user_id))
    }
    mx_session(
               server = server,
               token = resp$access_token,
               user_id = resp$user_id,
               device_id = resp$device_id
    )
}

#' Log in to a Matrix homeserver
#'
#' Authenticates with a Matrix homeserver using password login and returns
#' a session object carrying the access token and device id.
#'
#' @param server Character. Homeserver base URL (e.g. "https://matrix.example").
#' @param user Character. User localpart or full Matrix ID.
#' @param password Character. Account password.
#' @param device_id Character or NULL. Reuse an existing device id.
#'
#' @return An object of class "mx_session".
#' @examples
#' \dontrun{
#' s <- mx_login("https://matrix.example", "alice", "hunter2")
#' }
#' @export
mx_login <- function(server, user, password, device_id = NULL) {
    identifier <- if (grepl("^@", user)) {
        list(type = "m.id.user", user = sub("^@", "", sub(":.*$", "", user)))
    } else {
        list(type = "m.id.user", user = user)
    }
    body <- list(
                 type = "m.login.password",
                 identifier = identifier,
                 password = password
    )
    if (!is.null(device_id)) {
        body$device_id <- device_id
    }

    resp <- mx_http(server, "POST", "/_matrix/client/v3/login", body = body)
    mx_session(
               server = server,
               token = resp$access_token,
               user_id = resp$user_id,
               device_id = resp$device_id
    )
}

#' Reconstruct a session from saved credentials
#'
#' @param server Character. Homeserver base URL.
#' @param token Character. Access token from a prior login.
#' @param user_id Character. Full Matrix ID (e.g. "@troy:example.org").
#' @param device_id Character. Device id from the prior login.
#'
#' @return An object of class "mx_session".
#' @examples
#' s <- mx_session(
#'     server = "https://matrix.example",
#'     token = "syt_...",
#'     user_id = "@alice:matrix.example",
#'     device_id = "ABC123"
#' )
#' @export
mx_session <- function(server, token, user_id, device_id) {
    structure(
              list(
                   server = sub("/$", "", server),
                   token = token,
                   user_id = user_id,
                   device_id = device_id
        ),
              class = "mx_session"
    )
}

#' Log out of a Matrix session
#'
#' Invalidates the access token on the homeserver.
#'
#' @param session An "mx_session" object.
#'
#' @return Invisible NULL.
#' @examples
#' \dontrun{
#' mx_logout(s)
#' }
#' @export
mx_logout <- function(session) {
    mx_http(
            session$server, "POST", "/_matrix/client/v3/logout",
            body = mx_empty_body(), token = session$token
    )
    invisible(NULL)
}

#' Return the identity of the current session
#'
#' @param session An "mx_session" object.
#'
#' @return A list with user_id and device_id.
#' @examples
#' \dontrun{
#' mx_whoami(s)
#' }
#' @export
mx_whoami <- function(session) {
    resp <- mx_http(
                    session$server, "GET", "/_matrix/client/v3/account/whoami",
                    token = session$token
    )
    list(user_id = resp$user_id, device_id = resp$device_id)
}

library(tinytest)

# All live-server tests are gated on at_home() so R CMD check sees no
# network. Locally, set MX_TEST_SERVER / MX_TEST_USER / MX_TEST_PASS to
# run against a real Synapse.

expect_true(is.function(mx.api::mx_login))
expect_true(is.function(mx.api::mx_session))
expect_true(is.function(mx.api::mx_logout))
expect_true(is.function(mx.api::mx_whoami))
expect_true(is.function(mx.api::mx_register))

s <- mx.api::mx_session(
  server = "https://example/",
  token = "tok",
  user_id = "@u:example",
  device_id = "DEV"
)
expect_inherits(s, "mx_session")
expect_equal(s$server, "https://example")
expect_equal(s$token, "tok")

if (at_home() && nzchar(Sys.getenv("MX_TEST_SERVER"))) {
  s <- mx.api::mx_login(
    Sys.getenv("MX_TEST_SERVER"),
    Sys.getenv("MX_TEST_USER"),
    Sys.getenv("MX_TEST_PASS")
  )
  expect_inherits(s, "mx_session")
  wid <- mx.api::mx_whoami(s)
  expect_equal(wid$user_id, s$user_id)
  mx.api::mx_logout(s)
}

library(tinytest)

expect_true(is.function(mx.api::mx_rooms))
expect_true(is.function(mx.api::mx_room_create))
expect_true(is.function(mx.api::mx_room_join))
expect_true(is.function(mx.api::mx_room_leave))
expect_true(is.function(mx.api::mx_room_members))
expect_true(is.function(mx.api::mx_room_name))
expect_true(is.function(mx.api::mx_room_topic))

if (at_home() && nzchar(Sys.getenv("MX_TEST_SERVER"))) {
  # live room create/join/leave round-trip goes here
}

library(tinytest)

# Smoke: profile, invite, redact, typing signatures.
expect_true(is.function(mx.api::mx_profile))
expect_true(is.function(mx.api::mx_set_displayname))
expect_true(is.function(mx.api::mx_set_avatar_url))
expect_true(is.function(mx.api::mx_room_invite))
expect_true(is.function(mx.api::mx_redact))
expect_true(is.function(mx.api::mx_typing))

expect_equal(names(formals(mx.api::mx_profile)), c("session", "user_id"))
expect_equal(
  names(formals(mx.api::mx_redact)),
  c("session", "room_id", "event_id", "reason", "txn_id")
)
expect_equal(
  names(formals(mx.api::mx_typing)),
  c("session", "room_id", "typing", "timeout")
)
expect_equal(
  names(formals(mx.api::mx_room_invite)),
  c("session", "room_id", "user_id")
)

library(tinytest)

expect_true(is.function(mx.api::mx_devices))
expect_true(is.function(mx.api::mx_delete_device))
expect_equal(names(formals(mx.api::mx_devices)), "session")
expect_equal(
  names(formals(mx.api::mx_delete_device)),
  c("session", "device_id", "auth")
)

library(tinytest)

# Smoke: functions exist with the expected signatures.
expect_true(is.function(mx.api::mx_get_state))
expect_true(is.function(mx.api::mx_set_state))
expect_true(is.function(mx.api::mx_get_account_data))
expect_true(is.function(mx.api::mx_set_account_data))

expect_equal(
  names(formals(mx.api::mx_get_state)),
  c("session", "room_id", "event_type", "state_key")
)
expect_equal(
  names(formals(mx.api::mx_set_state)),
  c("session", "room_id", "event_type", "content", "state_key")
)
expect_equal(
  names(formals(mx.api::mx_get_account_data)),
  c("session", "type", "user_id")
)
expect_equal(
  names(formals(mx.api::mx_set_account_data)),
  c("session", "type", "content", "user_id")
)

if (at_home() && nzchar(Sys.getenv("MX_TEST_SERVER"))) {
  # live get/set round-trips go here
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_canonical_json}
\alias{mx_canonical_json}
\title{Encode a value as Matrix canonical JSON}
\usage{
mx_canonical_json(x)
}
\arguments{
\item{x}{An R value: NULL, atomic vector, or list.}
}
\value{
A length-1 UTF-8 character string. The result is the exact
  byte sequence to write to disk or feed to an ed25519 signer;
  non-ASCII content is preserved as raw UTF-8 bytes
  (jsonlite-style \\uXXXX escaping is not used).
}
\description{
Produces the canonical JSON byte sequence the Matrix specification
requires for signed objects: object keys sorted by UTF-8 byte
sequence, no insignificant whitespace, raw UTF-8 for non-ASCII
strings, integers only (no floats, no exponents, no decimal places,
within \code{[-(2^53)+1, (2^53)-1]}), and rejection of NaN, Inf,
NA values, and NA or duplicate object keys. The output is the exact
byte sequence to feed into an ed25519 signer.
}
\details{
R named lists become JSON objects; unnamed lists and length > 1
atomic vectors become arrays. Length-1 atomics become scalars. To
force a length-1 value to encode as an array, wrap it in a
single-element \code{list(...)} or in \code{I()} (AsIs values are
always encoded as arrays, names dropped, mirroring jsonlite).

}
\examples{
mx_canonical_json(list(b = 2, a = 1))
# "{\"a\":1,\"b\":2}"

mx_canonical_json(list(key = "abc"))
# "{\"key\":\"abc\"}"

}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_set_state}
\alias{mx_set_state}
\title{Set a room state event}
\usage{
mx_set_state(session, room_id, event_type, content, state_key = "")
}
\arguments{
\item{session}{An "mx_session" object.}

\item{room_id}{Character. The room ID.}

\item{event_type}{Character. State event type, e.g.
\code{"m.room.encryption"}.}

\item{content}{List. The state content, sent as-is.}

\item{state_key}{Character. State key (default empty string).}
}
\value{
The event ID of the state event.
}
\description{
Generic state setter, e.g. to mark a room encrypted by putting an
\code{m.room.encryption} event.
}
\examples{
\dontrun{
mx_set_state(s, "!abc:example", "m.room.encryption",
             list(algorithm = "m.megolm.v1.aes-sha2"))
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_rooms}
\alias{mx_rooms}
\title{List rooms the user has joined}
\usage{
mx_rooms(session)
}
\arguments{
\item{session}{An "mx_session" object.}
}
\value{
Character vector of room IDs.
}
\description{
List rooms the user has joined
}
\examples{
\dontrun{
mx_rooms(s)
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_guess_mime}
\alias{mx_guess_mime}
\title{Guess a MIME type from a file extension}
\usage{
mx_guess_mime(path)
}
\arguments{
\item{path}{Character. File path or name.}
}
\value{
Character MIME type.
}
\description{
The extension table mx.api uses for uploads, exported so callers do
not maintain their own. Unknown extensions fall back to
\code{"application/octet-stream"}.
}
\examples{
mx_guess_mime("clip.mp4")
mx_guess_mime("notes.txt")
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_set_displayname}
\alias{mx_set_displayname}
\title{Set this user's display name}
\usage{
mx_set_displayname(session, displayname)
}
\arguments{
\item{session}{An "mx_session" object.}

\item{displayname}{Character. The new display name.}
}
\value{
Invisibly TRUE on success.
}
\description{
Set this user's display name
}
\examples{
\dontrun{
mx_set_displayname(s, "cornelius")
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_react}
\alias{mx_react}
\title{Send a reaction (annotation) to a room event}
\usage{
mx_react(session, room_id, event_id, key)
}
\arguments{
\item{session}{An "mx_session" object.}

\item{room_id}{Character. The room ID.}

\item{event_id}{Character. The event being reacted to.}

\item{key}{Character. The reaction key (usually an emoji).}
}
\value{
The event ID of the sent reaction.
}
\description{
Posts an m.reaction event tying \code{key} (usually a thumbs-up or
other emoji) to \code{event_id}. Matrix reactions are plain events
under the hood; they relate to the target via \code{m.annotation}.
}
\examples{
\dontrun{
mx_react(s, "!abc:matrix.example", "$eventid", "thumbs-up")
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_send}
\alias{mx_send}
\title{Send a message to a room}
\usage{
mx_send(session, room_id, body, msgtype = "m.text", extra = NULL)
}
\arguments{
\item{session}{An "mx_session" object.}

\item{room_id}{Character. The room ID.}

\item{body}{Character. The message body.}

\item{msgtype}{Character. Matrix msgtype, default "m.text".}

\item{extra}{List or NULL. Extra fields merged into the event content
(e.g. formatted body, reply relation).}
}
\value{
The event ID of the sent message.
}
\description{
Send a message to a room
}
\examples{
\dontrun{
mx_send(s, "!abc:matrix.example", "hello world")
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_room_leave}
\alias{mx_room_leave}
\title{Leave a room}
\usage{
mx_room_leave(session, room_id)
}
\arguments{
\item{session}{An "mx_session" object.}

\item{room_id}{Character. The room ID.}
}
\value{
Invisible NULL.
}
\description{
Leave a room
}
\examples{
\dontrun{
mx_room_leave(s, "!abc:matrix.example")
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_set_avatar_url}
\alias{mx_set_avatar_url}
\title{Set this user's avatar}
\usage{
mx_set_avatar_url(session, avatar_url)
}
\arguments{
\item{session}{An "mx_session" object.}

\item{avatar_url}{Character. An \code{mxc://} URI, typically from
\code{\link{mx_upload}}.}
}
\value{
Invisibly TRUE on success.
}
\description{
Set this user's avatar
}
\examples{
\dontrun{
uri <- mx_upload(s, "avatar.png")
mx_set_avatar_url(s, uri)
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_profile}
\alias{mx_profile}
\title{Get a user's profile}
\usage{
mx_profile(session, user_id = session$user_id)
}
\arguments{
\item{session}{An "mx_session" object.}

\item{user_id}{Character. Defaults to the session's own user.}
}
\value{
A list with \code{displayname} and \code{avatar_url} (either
  may be absent when unset).
}
\description{
Get a user's profile
}
\examples{
\dontrun{
mx_profile(s)$displayname
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_sync}
\alias{mx_sync}
\title{One-shot sync against the homeserver}
\usage{
mx_sync(session, since = NULL, timeout = 0L, filter = NULL)
}
\arguments{
\item{session}{An "mx_session" object.}

\item{since}{Character or NULL. Sync token from a prior sync.}

\item{timeout}{Integer. Long-poll timeout in milliseconds (0 returns
immediately).}

\item{filter}{Character or NULL. Filter ID or inline JSON filter.}
}
\value{
The parsed sync response, including next_batch.
}
\description{
Calls /sync once and returns immediately. For streaming behaviour,
the caller writes its own loop, passing the previous batch's
next_batch token as \code{since}.
}
\examples{
\dontrun{
batch <- mx_sync(s)
next_batch <- batch$next_batch
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_keys_claim}
\alias{mx_keys_claim}
\title{Claim one-time keys for an Olm handshake}
\usage{
mx_keys_claim(session, one_time_keys, timeout = 10000L)
}
\arguments{
\item{session}{An \code{mx_session}.}

\item{one_time_keys}{Named list. Names are user ids; values are
named lists mapping device ids to the desired algorithm
(typically \code{"signed_curve25519"}).}

\item{timeout}{Integer milliseconds. Server-side timeout when
talking to remote homeservers.}
}
\value{
Parsed response with the claimed keys keyed by
  \code{user_id -> device_id -> "<algorithm>:<key_id>" -> key_object}.
}
\description{
POST \code{/_matrix/client/v3/keys/claim}. The
\code{one_time_keys} argument selects which algorithm to claim for
each \code{(user_id, device_id)} pair.
}
\examples{
\dontrun{
mx_keys_claim(s, list(
  "@alice:example.org" = list(ABCD1234 = "signed_curve25519")
))
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_room_create}
\alias{mx_room_create}
\title{Create a room}
\usage{
mx_room_create(session, name = NULL, topic = NULL, visibility = "private",
               preset = NULL, invite = character())
}
\arguments{
\item{session}{An "mx_session" object.}

\item{name}{Character or NULL. Human-readable room name.}

\item{topic}{Character or NULL. Room topic.}

\item{visibility}{Character. "private" (default) or "public".}

\item{preset}{Character or NULL. A Matrix room preset
("private_chat", "trusted_private_chat", "public_chat").}

\item{invite}{Character vector. Matrix IDs to invite.}
}
\value{
The new room ID as a character string.
}
\description{
Create a room
}
\examples{
\dontrun{
room_id <- mx_room_create(s, name = "test", topic = "hello")
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_read_receipt}
\alias{mx_read_receipt}
\title{Send a read receipt for a room event}
\usage{
mx_read_receipt(session, room_id, event_id,
                receipt_type = c("m.read", "m.read.private"))
}
\arguments{
\item{session}{An "mx_session" object.}

\item{room_id}{Character. The room ID.}

\item{event_id}{Character. The event to mark as read.}

\item{receipt_type}{Character. "m.read" (default) or "m.read.private".}
}
\value{
Invisible NULL.
}
\description{
Public receipt (\code{m.read}) advances the "seen" marker in other
clients; private receipt (\code{m.read.private}) only advances the
bot's own view. Defaults to public so user clients show
"seen by @bot".
}
\examples{
\dontrun{
mx_read_receipt(s, "!abc:matrix.example", "$eventid")
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\docType{package}
\name{mx.api-package}
\alias{mx.api}
\alias{mx.api-package}
\title{\packageTitle{mx.api}}
\description{\packageDescription{mx.api}}
\section{Package Content}{\packageIndices{mx.api}}
\author{\packageAuthor{mx.api}}
\section{Maintainer}{\packageMaintainer{mx.api}}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_send_to_device}
\alias{mx_send_to_device}
\title{Send a to-device event}
\usage{
mx_send_to_device(session, event_type, messages, txn_id = NULL)
}
\arguments{
\item{session}{An \code{mx_session}.}

\item{event_type}{Character. The to-device event type, e.g.
\code{"m.room.encrypted"}.}

\item{messages}{Named list. Outer names are user ids; values are
named lists mapping device id (or the wildcard \code{"*"}) to
the event content.}

\item{txn_id}{Character or NULL. Idempotency key; auto-generated
when NULL.}
}
\value{
Invisible NULL (the server returns an empty body on success).
}
\description{
PUT \code{/_matrix/client/v3/sendToDevice/{eventType}/{txnId}}.
Used to ship encrypted Olm payloads (e.g. m.room_key carriers
wrapped as \code{m.room.encrypted}) to specific
\code{(user_id, device_id)} targets.
}
\examples{
\dontrun{
mx_send_to_device(s, "m.room.encrypted", list(
  "@bob:example.org" = list(BBBB = encrypted_content)
))
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_set_account_data}
\alias{mx_set_account_data}
\title{Set account data}
\usage{
mx_set_account_data(session, type, content, user_id = session$user_id)
}
\arguments{
\item{session}{An "mx_session" object.}

\item{type}{Character. Event type, e.g. \code{"m.direct"}.}

\item{content}{List. The content, sent as-is.}

\item{user_id}{Character. Defaults to the session's own user.}
}
\value{
Invisibly TRUE on success.
}
\description{
Writes a global account-data event for a user. The content replaces
whatever was stored under \code{type}.
}
\examples{
\dontrun{
mx_set_account_data(s, "ai.cornball.notes", list(theme = "dark"))
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_session}
\alias{mx_session}
\title{Reconstruct a session from saved credentials}
\usage{
mx_session(server, token, user_id, device_id)
}
\arguments{
\item{server}{Character. Homeserver base URL.}

\item{token}{Character. Access token from a prior login.}

\item{user_id}{Character. Full Matrix ID (e.g. "@troy:example.org").}

\item{device_id}{Character. Device id from the prior login.}
}
\value{
An object of class "mx_session".
}
\description{
Reconstruct a session from saved credentials
}
\examples{
s <- mx_session(
    server = "https://matrix.example",
    token = "syt_...",
    user_id = "@alice:matrix.example",
    device_id = "ABC123"
)
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_messages}
\alias{mx_messages}
\title{Fetch historical messages from a room}
\usage{
mx_messages(session, room_id, from = NULL, dir = "b", limit = 50L)
}
\arguments{
\item{session}{An "mx_session" object.}

\item{room_id}{Character. The room ID.}

\item{from}{Character or NULL. Pagination token; NULL starts at the
most recent message.}

\item{dir}{Character. "b" (backwards, default) or "f" (forwards).}

\item{limit}{Integer. Maximum events to return.}
}
\value{
A list with fields chunk, start, end.
}
\description{
Thin wrapper over the /rooms/\{id\}/messages endpoint.
}
\examples{
\dontrun{
mx_messages(s, "!abc:matrix.example", limit = 20L)
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_whoami}
\alias{mx_whoami}
\title{Return the identity of the current session}
\usage{
mx_whoami(session)
}
\arguments{
\item{session}{An "mx_session" object.}
}
\value{
A list with user_id and device_id.
}
\description{
Return the identity of the current session
}
\examples{
\dontrun{
mx_whoami(s)
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_keys_upload}
\alias{mx_keys_upload}
\title{Upload device identity and one-time keys}
\usage{
mx_keys_upload(session, device_keys = NULL, one_time_keys = NULL,
               fallback_keys = NULL)
}
\arguments{
\item{session}{An \code{mx_session}.}

\item{device_keys}{Named list. The \code{device_keys} object as
defined in the Matrix spec, including \code{user_id},
\code{device_id}, \code{algorithms}, \code{keys}, and the
\code{signatures} block produced by the caller's signer. Pass
NULL to upload only one-time keys.}

\item{one_time_keys}{Named list or NULL. Map from
\code{"<algorithm>:<key_id>"} (e.g. \code{"signed_curve25519:AAAA"})
to the signed key object. Pass NULL or an empty list to skip.}

\item{fallback_keys}{Named list or NULL. Same shape as
\code{one_time_keys}; used to advertise a fallback key when the
OTK pool is exhausted.}
}
\value{
The parsed homeserver response, including
  \code{one_time_key_counts}.
}
\description{
POST \code{/_matrix/client/v3/keys/upload}. The \code{device_keys}
and \code{one_time_keys} arguments must be fully formed and signed
per the Matrix specification; mx.api will not canonicalise or sign
them. Use \code{\link{mx_canonical_json}} to produce the byte
sequence to sign.
}
\examples{
\dontrun{
mx_keys_upload(s, device_keys = signed_dk, one_time_keys = signed_otks)
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_login}
\alias{mx_login}
\title{Log in to a Matrix homeserver}
\usage{
mx_login(server, user, password, device_id = NULL)
}
\arguments{
\item{server}{Character. Homeserver base URL (e.g. "https://matrix.example").}

\item{user}{Character. User localpart or full Matrix ID.}

\item{password}{Character. Account password.}

\item{device_id}{Character or NULL. Reuse an existing device id.}
}
\value{
An object of class "mx_session".
}
\description{
Authenticates with a Matrix homeserver using password login and returns
a session object carrying the access token and device id.
}
\examples{
\dontrun{
s <- mx_login("https://matrix.example", "alice", "hunter2")
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_get_state}
\alias{mx_get_state}
\title{Get a room state event}
\usage{
mx_get_state(session, room_id, event_type, state_key = "")
}
\arguments{
\item{session}{An "mx_session" object.}

\item{room_id}{Character. The room ID.}

\item{event_type}{Character. State event type, e.g.
\code{"m.room.encryption"}.}

\item{state_key}{Character. State key (default empty string).}
}
\value{
The state event content as a list, or NULL when the state
  event is not set.
}
\description{
Read-side counterpart of \code{\link{mx_set_state}}, e.g. to check
whether a room is encrypted before joining the send path.
}
\examples{
\dontrun{
enc <- mx_get_state(s, "!abc:example", "m.room.encryption")
is.null(enc)   # FALSE in an encrypted room
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_room_invite}
\alias{mx_room_invite}
\title{Invite a user to a room}
\usage{
mx_room_invite(session, room_id, user_id)
}
\arguments{
\item{session}{An "mx_session" object.}

\item{room_id}{Character. The room ID.}

\item{user_id}{Character. The Matrix ID to invite.}
}
\value{
Invisibly TRUE on success.
}
\description{
Invitation at creation time is covered by \code{mx_room_create()};
this covers the other common lifecycle case, inviting into an
existing room.
}
\examples{
\dontrun{
mx_room_invite(s, "!abc:example", "@friend:example.org")
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_upload}
\alias{mx_upload}
\title{Upload a file to the homeserver media repository}
\usage{
mx_upload(session, path, content_type = NULL, filename = NULL)
}
\arguments{
\item{session}{An "mx_session" object.}

\item{path}{Character. Local file path.}

\item{content_type}{Character or NULL. MIME type; guessed from the
file extension if NULL.}

\item{filename}{Character or NULL. Filename advertised to the server.}
}
\value{
An "mxc://" URI as a character string.
}
\description{
Upload a file to the homeserver media repository
}
\examples{
\dontrun{
uri <- mx_upload(s, "photo.png")
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_download}
\alias{mx_download}
\title{Download a media file by mxc URI}
\usage{
mx_download(session, mxc_url, dest)
}
\arguments{
\item{session}{An "mx_session" object.}

\item{mxc_url}{Character. An "mxc://server/id" URI.}

\item{dest}{Character. Destination file path.}
}
\value{
The destination path, invisibly.
}
\description{
Download a media file by mxc URI
}
\examples{
\dontrun{
mx_download(s, "mxc://matrix.example/abc123", tempfile())
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_devices}
\alias{mx_devices}
\title{List this account's devices}
\usage{
mx_devices(session)
}
\arguments{
\item{session}{An "mx_session" object.}
}
\value{
A list of devices, each with \code{device_id},
  \code{display_name}, \code{last_seen_ip}, and \code{last_seen_ts}.
}
\description{
List this account's devices
}
\examples{
\dontrun{
vapply(mx_devices(s), function(d) d$device_id, character(1))
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_room_join}
\alias{mx_room_join}
\title{Join a room by ID or alias}
\usage{
mx_room_join(session, room)
}
\arguments{
\item{session}{An "mx_session" object.}

\item{room}{Character. Room ID (!abc:server) or alias (#name:server).}
}
\value{
The joined room ID.
}
\description{
Join a room by ID or alias
}
\examples{
\dontrun{
mx_room_join(s, "#general:matrix.example")
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_logout}
\alias{mx_logout}
\title{Log out of a Matrix session}
\usage{
mx_logout(session)
}
\arguments{
\item{session}{An "mx_session" object.}
}
\value{
Invisible NULL.
}
\description{
Invalidates the access token on the homeserver.
}
\examples{
\dontrun{
mx_logout(s)
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_typing}
\alias{mx_typing}
\title{Send a typing notification}
\usage{
mx_typing(session, room_id, typing = TRUE, timeout = 30000L)
}
\arguments{
\item{session}{An "mx_session" object.}

\item{room_id}{Character. The room ID.}

\item{typing}{Logical. TRUE to show typing, FALSE to clear it.}

\item{timeout}{Integer. How long the indicator lasts, in
milliseconds (ignored when \code{typing = FALSE}).}
}
\value{
Invisibly TRUE on success.
}
\description{
Shows (or clears) the session user's typing indicator in a room.
Useful bot polish while a slow reply is being generated.
}
\examples{
\dontrun{
mx_typing(s, "!abc:example", TRUE)
# ... generate the reply ...
mx_typing(s, "!abc:example", FALSE)
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_media_config}
\alias{mx_media_config}
\title{Query the homeserver's media configuration}
\usage{
mx_media_config(session)
}
\arguments{
\item{session}{An "mx_session" object.}
}
\value{
A list; \code{$`m.upload.size`} is the upload cap in bytes
  (may be absent if the server does not advertise one).
}
\description{
Asks the server for its media limits, chiefly \code{m.upload.size}
(maximum upload bytes), so callers can check a file fits before
uploading. Tries the v1 endpoint and falls back to the legacy
location for older homeservers.
}
\examples{
\dontrun{
cap <- mx_media_config(s)$`m.upload.size`
file.size("clip.mp4") <= cap
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_keys_query}
\alias{mx_keys_query}
\title{Query device keys for one or more users}
\usage{
mx_keys_query(session, device_keys, timeout = 10000L)
}
\arguments{
\item{session}{An \code{mx_session}.}

\item{device_keys}{Named list. Names are Matrix user ids
(e.g. \code{"@@alice:example.org"}); values are character vectors
of device ids, or \code{character(0)} for "all devices".}

\item{timeout}{Integer milliseconds. Time the server should wait
for remote homeservers before returning a partial result.}
}
\value{
Parsed response with a \code{device_keys} map of
  \code{user_id -> device_id -> device_keys_object}.
}
\description{
POST \code{/_matrix/client/v3/keys/query}. Each entry in
\code{device_keys} maps a Matrix user id to a character vector of
device ids to request. An empty character vector requests all of
the user's devices.
}
\examples{
\dontrun{
mx_keys_query(s, list("@alice:example.org" = character()))
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_room_topic}
\alias{mx_room_topic}
\title{Get a room's topic}
\usage{
mx_room_topic(session, room_id)
}
\arguments{
\item{session}{An "mx_session" object.}

\item{room_id}{Character. The room ID.}
}
\value{
Character scalar or NULL.
}
\description{
Reads the \code{m.room.topic} state event. Returns NULL if the room
has no topic set or the state event is inaccessible.
}
\examples{
\dontrun{
mx_room_topic(s, "!abc:matrix.example")
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_delete_device}
\alias{mx_delete_device}
\title{Delete a device}
\usage{
mx_delete_device(session, device_id, auth = NULL)
}
\arguments{
\item{session}{An "mx_session" object.}

\item{device_id}{Character. The device to delete.}

\item{auth}{List or NULL. Completed user-interactive auth payload.}
}
\value{
Invisibly TRUE on success.
}
\description{
Removes a device and invalidates its access token. Most homeservers
protect this with user-interactive authentication: the first call
fails with M_FORBIDDEN or a 401 carrying a \code{flows} object, and
the caller retries with a completed \code{auth} payload, e.g.
\code{list(type = "m.login.password", identifier = list(type =
"m.id.user", user = "bot"), password = "...", session = "<from the
401>")}. mx.api deliberately does not automate that exchange.
}
\examples{
\dontrun{
mx_delete_device(s, "OLDDEVICE", auth = list(
    type = "m.login.password",
    identifier = list(type = "m.id.user", user = "bot"),
    password = "secret"
))
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_register}
\alias{mx_register}
\title{Register a new account on a Matrix homeserver}
\usage{
mx_register(server, username, password, device_id = NULL,
            initial_device_display_name = NULL, inhibit_login = FALSE)
}
\arguments{
\item{server}{Character. Homeserver base URL.}

\item{username}{Character. Desired localpart (e.g. "alice").}

\item{password}{Character. Account password.}

\item{device_id}{Character or NULL. Device id to assign; a server-
generated one is used if NULL.}

\item{initial_device_display_name}{Character or NULL. Human-readable
label for the device.}

\item{inhibit_login}{Logical. When TRUE, the server creates the
account but does not return a session; the call returns a list
with the new \code{user_id} instead of an \code{mx_session}.}
}
\value{
An \code{mx_session} object on login, or a list with
  \code{user_id} when \code{inhibit_login = TRUE}.
}
\description{
Creates a new user via POST \code{/_matrix/client/v3/register} using
the \code{m.login.dummy} auth flow. Most homeservers only accept this
when open registration is enabled (or a registration token is
supplied). On success returns a ready-to-use \code{mx_session} —
registration also logs the new user in.
}
\examples{
\dontrun{
s <- mx_register("https://matrix.example", "alice", "hunter2")
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_redact}
\alias{mx_redact}
\title{Redact an event}
\usage{
mx_redact(session, room_id, event_id, reason = NULL, txn_id = NULL)
}
\arguments{
\item{session}{An "mx_session" object.}

\item{room_id}{Character. The room ID.}

\item{event_id}{Character. The event to redact.}

\item{reason}{Character or NULL. Optional human-readable reason.}

\item{txn_id}{Character or NULL. Transaction id (generated if NULL).}
}
\value{
The event ID of the redaction event.
}
\description{
Removes the content of a message, reaction, or other event. This is
how Matrix deletes things.
}
\examples{
\dontrun{
mx_redact(s, "!abc:example", "$someevent", reason = "typo")
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_send_event}
\alias{mx_send_event}
\title{Send an arbitrary room event}
\usage{
mx_send_event(session, room_id, event_type, content, txn_id = NULL)
}
\arguments{
\item{session}{An "mx_session" object.}

\item{room_id}{Character. The room ID.}

\item{event_type}{Character. The event type, e.g.
\code{"m.room.encrypted"}.}

\item{content}{List. The event content, sent as-is.}

\item{txn_id}{Character or NULL. Transaction id (generated if NULL).}
}
\value{
The event ID of the sent event.
}
\description{
Generic counterpart to \code{\link{mx_send}} for event types other than
\code{m.room.message}, such as \code{m.room.encrypted}. The content is
sent verbatim.
}
\examples{
\dontrun{
mx_send_event(s, "!abc:example", "m.room.encrypted", encrypted_content)
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_send_media}
\alias{mx_send_media}
\alias{mx_send_file}
\alias{mx_send_image}
\alias{mx_send_audio}
\alias{mx_send_video}
\title{Send a media file to a room}
\usage{
mx_send_media(session, room_id, path, body = basename(path), msgtype = NULL,
              content_type = NULL, info = list())

mx_send_file(session, room_id, path, body = basename(path),
             content_type = NULL, info = list())

mx_send_image(session, room_id, path, body = basename(path),
              content_type = NULL, info = list())

mx_send_audio(session, room_id, path, body = basename(path),
              content_type = NULL, info = list())

mx_send_video(session, room_id, path, body = basename(path),
              content_type = NULL, info = list())
}
\arguments{
\item{session}{An "mx_session" object.}

\item{room_id}{Character. The room ID.}

\item{path}{Character. Path to the file to upload.}

\item{body}{Character. Message body / filename shown by clients.}

\item{msgtype}{Character or NULL. One of \code{"m.file"},
\code{"m.image"}, \code{"m.audio"}, \code{"m.video"}. NULL (the
default) derives it from the MIME type, so a .mp4 posts as m.video
without being told.}

\item{content_type}{Character or NULL. MIME type (guessed from the
extension when NULL).}

\item{info}{List. Extra fields merged into the \code{info} object.}
}
\value{
The event ID of the sent message.
}
\description{
Uploads \code{path} to the media repository and posts an
\code{m.room.message} referencing it. The default \code{info} carries
\code{mimetype} and \code{size}; pass richer metadata (width, height,
duration) yourself -- mx.api deliberately does not inspect media files.
}
\examples{
\dontrun{
mx_send_media(s, "!abc:example", "clip.mp4", msgtype = "m.video")
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_get_account_data}
\alias{mx_get_account_data}
\title{Get account data}
\usage{
mx_get_account_data(session, type, user_id = session$user_id)
}
\arguments{
\item{session}{An "mx_session" object.}

\item{type}{Character. Event type, e.g. \code{"m.direct"}.}

\item{user_id}{Character. Defaults to the session's own user.}
}
\value{
The account-data content as a list, or NULL when the type has
  never been set.
}
\description{
Reads a global account-data event for a user, e.g. \code{"m.direct"}
(the DM room map) or any custom namespaced type.
}
\examples{
\dontrun{
direct <- mx_get_account_data(s, "m.direct")
}
}
